#!/usr/bin/env python3
"""Unit tests for log.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/14"

import unittest
import enb


class TestLog(unittest.TestCase):
    def test_verbose_updates_logging(self):
        original_log_level_name = enb.log.logger.selected_log_level.name
        original_verbosity_level = enb.config.options.verbose
        try:
            levels_by_priority = enb.log.logger.levels_by_priority()
            assert len(levels_by_priority) > 2

            for i in range(1, len(levels_by_priority)):
                assert levels_by_priority[i - 1].priority <= levels_by_priority[i].priority

            for i in range(len(levels_by_priority) - 1):
                for j in range(i + 1, len(levels_by_priority)):
                    enb.config.options.verbose = 0
                    enb.log.logger.selected_log_level = levels_by_priority[i]
                    enb.config.options.verbose = \
                        levels_by_priority[j].priority \
                        - enb.log.logger.level_message.priority
                    assert enb.log.logger.selected_log_level is levels_by_priority[j], \
                        dict(i=i, j=j, verbose=enb.config.options.verbose,
                             selected_level=enb.log.logger.selected_log_level,
                             target_level_j=levels_by_priority[j])

                    enb.config.options.verbose = 0
                    enb.log.logger.selected_log_level = levels_by_priority[j]
                    enb.config.options.verbose += \
                        levels_by_priority[i].priority \
                        - enb.log.logger.level_message.priority
                    assert enb.log.logger.selected_log_level is levels_by_priority[i], \
                        dict(i=i, j=j, verbose=enb.config.options.verbose,
                             selected_level=enb.log.logger.selected_log_level,
                             target_level_i=levels_by_priority[i])
        finally:
            enb.config.options.verbose = original_verbosity_level
            enb.log.logger.selected_log_level = enb.log.get_level(original_log_level_name)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Example showing the definition of a basic expeirment with `enb`.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/11/11"

import glob
import socket
import enb


class CheckPort(enb.experiment.ExperimentTask):
    """Experiment task that can perform a check based on
    the selected port.
    """

    def __init__(self, port):
        assert 0 <= port <= 65353
        super().__init__(param_dict=dict(port=port))

    def check_one_port(self, url_file_path):
        """Check whether a given port for a given url in url_file_path is open.
        Return True if and only if a socket could be open to that url:port.
        """
        with open(url_file_path, "r") as url_file:
            url = url_file.readline().strip()
        a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        location = (url, self.param_dict['port'])
        result_of_check = a_socket.connect_ex(location)
        return result_of_check == 0


class PortExperiment(enb.experiment.Experiment):
    """Example experiment that checks whether certain ports are open.
    """
    dataset_files_extension = "txt"

    def column_port_open(self, index, row):
        url_file_path, task_name = index
        checkport_task = self.tasks_by_name[task_name]
        return checkport_task.check_one_port(url_file_path)


if __name__ == '__main__':
    # This is the list of inputs. Each one contains one IP in a single line.
    enb.config.options.base_dataset_dir = "./data/ips"

    # This is the list of tasks to be run
    tasks = [CheckPort(port=p) for p in [53, 22, 80, 443, 8080]]

    # Obtain the full table of results
    exp = PortExperiment(tasks=tasks)
    result_df = exp.get_df()

    # Display a list of observed open ports
    open_port_df = result_df[result_df["port_open"]]
    for url_file_path, url_df in open_port_df.groupby("file_path"):
        print(f"Found open ports in {open(url_file_path, 'r').read().strip()}:\n - ", end="")
        print("\n - ".join(str(d['port']) for d in url_df["param_dict"]))
            



import enb.plugins


class LCNLPlugin(enb.plugins.Plugin):
    name = "ccsds124x0"
    label = "Wrappers for CCSDS 124.0-B-1"
    tags = {"data compression", "codec", "privative"}
    contrib_authors = ["Miguel Hernández-Cabronero", "Ian Blanes"]
    extra_requirements_message = "Binaries for CCSDS 124.0-B-1 are not available due to intellectual property issues."
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Codec wrapper for the zfp lossless image coder 
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
import enb


class Zfp(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.WrapperCodec):
    """Wrapper for the zfp codec
    Allowed data type to be compressed: integer 32 & 64 , float 32 & 64
    """

    def __init__(self, zfp_binary=os.path.join(os.path.dirname(__file__), "zfp")):
        """
        param:dtype: valid types in zfp are f32, f64, i32 and i64
        """
        super().__init__(compressor_path=zfp_binary,
                         decompressor_path=zfp_binary,
                         param_dict=dict())

    @property
    def label(self):
        return "ZFP"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] != 2, 'data type can not be 16 bpp'
        dimensions = 1

        if original_file_info.float == True:
            return f"-i {os.path.abspath(original_path)} " \
                   f"-z {os.path.abspath(compressed_path)} " \
                   f"-t f{original_file_info.bytes_per_sample * 8} " \
                   f"-{dimensions} {original_file_info.samples} -R"

        if original_file_info.float == False:
            return f"-i {os.path.abspath(original_path)} " \
                   f"-z {os.path.abspath(compressed_path)} " \
                   f"-t i{original_file_info.bytes_per_sample * 8} " \
                   f"-{dimensions} {original_file_info.samples} -R"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        dimensions = 1
        # TODO: include {original_file_info.samples} instead of defining a dummy variable (here and everywhere else)
        if original_file_info.float == True:
            return f" -z {os.path.abspath(compressed_path)} " \
                   f"-o {os.path.abspath(reconstructed_path)}  " \
                   f"-t f{original_file_info.bytes_per_sample * 8} " \
                   f"-{dimensions} {original_file_info.samples} -R"

        if original_file_info.float == False:
            return f" -z {os.path.abspath(compressed_path)} " \
                   f"-o {os.path.abspath(reconstructed_path)}  " \
                   f"-t i{original_file_info.bytes_per_sample * 8} " \
                   f"-{dimensions} {original_file_info.samples} -R"



#!/usr/bin/env python
"""
Wrapper for the VVC codec, using the reference implementation from

https://vcgit.hhi.fraunhofer.de/jvet/VVCSoftware_VTM/-/tree/master
"""
__author__ = "Natalia Blasco, Ester Jara, Artur Llabrés and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
import math
from enb import icompression


class VVC(icompression.WrapperCodec):
    """
    Base class for VVC coder
    """
    default_qp = 0

    def __init__(self, config_path, chroma_format="400", qp=None):
        """
        :param chroma_format: Specifies the chroma format used in the input file (only 400 supported).
        :param qp: Specifies the base value of the quantization parameter (0-63, extended from HEVC's 0-51).
        """
        chroma_format = str(chroma_format)
        assert chroma_format in ["400"], f"Chroma format {chroma_format} not supported."
        qp = int(qp) if qp is not None else self.default_qp
        assert 0 <= qp <= 63
        self.config_path = config_path
        assert os.path.isfile(config_path), f"VVC configuration file path {repr(config_path)} not found."

        param_dict = dict(chroma_format=chroma_format, QP=qp, config_path=config_path)
        icompression.WrapperCodec.__init__(
            self,
            compressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "EncoderAppStatic"),
            decompressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "DecoderAppStatic"),
            param_dict=param_dict)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info['component_count'] == 1, f"Only 1 frame supported"
        return f"-i {original_path} " \
               f"-c {self.param_dict['config_path']} " \
               f"-b {compressed_path} " \
               f"-wdt {original_file_info['width']} " \
               f"-hgt {original_file_info['height']} " \
               f"-f {original_file_info['component_count']} " \
               f"-cf {self.param_dict['chroma_format']} " \
               f"--InputChromaFormat={self.param_dict['chroma_format']} " \
               f"--InputBitDepth={8 * original_file_info['bytes_per_sample']} " \
               f"--InternalBitDepth={8 * original_file_info['bytes_per_sample']} " \
               f"-fr 1 "

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert original_file_info['component_count'] == 1, f"Only 1 frame supported"
        return f"-b {compressed_path} " \
               f"-o {reconstructed_path} " \
               f"-d {8 * original_file_info['bytes_per_sample']}"

    @property
    def label(self):
        raise NotImplementedError("Subclasses must implement their own label.")


class VVC_lossless(VVC, icompression.LosslessCodec):
    """
    VVC subclass for lossless compression
    """

    def __init__(self, config_path=None, chroma_format="400"):
        """
        :param chroma_format: Specifies the chroma format used in the input file (only 400 supported).
        """
        config_path = config_path if config_path is not None \
            else os.path.join(os.path.dirname(os.path.abspath(__file__)),
                              f"vvc_lossless_{chroma_format}.cfg")
        VVC.__init__(self, config_path, chroma_format, qp=0)

    @property
    def label(self):
        return "VVC lossless"


class VVC_lossy(VVC, icompression.LossyCodec):
    """
    VVC subclass for lossy compression
    """
    rate_decimals = 3

    def __init__(self, config_path=None, chroma_format="400", qp=25, bit_rate=None):
        """
        :param chroma_format: Specifies the chroma format used in the input file (only 400 supported).
        :param qp: Specifies the base value of the quantization parameter (0-51).
        :param bit_rate: target bitrate in bps.
        """

        bit_rate = round(float(bit_rate), self.rate_decimals) if bit_rate is not None else bit_rate

        config_path = config_path if config_path is not None \
            else os.path.join(os.path.dirname(os.path.abspath(__file__)),
                              f"vvc_lossy_{chroma_format}.cfg")
        VVC.__init__(
            self, config_path=config_path, chroma_format=chroma_format, qp=qp)

        self.param_dict['bit_rate'] = bit_rate

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert self.param_dict['bit_rate'] is None, f"Bit rate not implemented"

        if original_file_info['bytes_per_sample'] > 1:
            raise Exception(f"Bytes per sample = {original_file_info['bytes_per_sample']} "
                            f"not supported yet.")

        params = super().get_compression_params(
            original_path=original_path,
            compressed_path=compressed_path,
            original_file_info=original_file_info)

        if self.param_dict['bit_rate'] is not None:
            target_rate = math.ceil(self.param_dict['bit_rate']
                                    * original_file_info['width'] * original_file_info['height'])
            params += " --RateControl"
            params += f" --TargetBitrate={target_rate} "
            # params += "--MaxSubLayers=16"
        else:
            params += f" --QP={self.param_dict['QP']} --InitialQP={self.param_dict['QP']}"
        return params

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert self.param_dict['bit_rate'] is None, f"Bit rate not implemented"

        if original_file_info['bytes_per_sample'] > 1:
            raise Exception(f"Bytes per sample = "
                            f"{original_file_info['bytes_per_sample']} "
                            f"not supported")
        else:
            return super().get_decompression_params(compressed_path=compressed_path,
                                                    reconstructed_path=reconstructed_path,
                                                    original_file_info=original_file_info)

    @property
    def label(self):
        if self.param_dict["bit_rate"] is not None:
            rate_format = "{0:." + str(self.rate_decimals) + "f}"
            s = f"VVC {rate_format.format(self.param_dict['bit_rate'])}bps"
        else:
            s = f"VVC QP{self.param_dict['QP']}"
        return s



#!/usr/bin/env python3
"""Codec wrapper for the Fpack lossless and lossy image coder 
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/07/01"

import os
import enb


class Fpack(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.FITSWrapperCodec):
    """Wrapper for the imageMarlin codec.
    """

    def __init__(self,
                 bin_dir=None,
                 quantization=0,
                 scale=0,
                 compression_method='r'):
        """
        :param bin_dir: path to the dir with the fapec and unfapec binaries
        :param compression method: compression method used (r - Rice, h - Hcompress,
        p - PLIO, g - Gzip or i2f - converts integer to floats and uses Rice compression )
        :param quanization: quantization level
        :param scale: Scale factor for lossy compression when using Hcompress        
        """

        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fpack"),
                         decompressor_path=os.path.join(bin_dir, "funpack"),
                         param_dict=dict(compression_method=compression_method, quantization=quantization, scale=scale))

    @property
    def name(self):
        """Don't include the binary signature
        """
        return f"{self.__class__.__name__}{'__' if self.param_dict else ''}" \
               f"{'_'.join(f'{k}={v}' for k, v in self.param_dict.items())}"

    @property
    def label(self):
        return "FPACK"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f" -{self.param_dict['compression_method']} " + \
               (f" -q {self.param_dict['quantization']} " if self.param_dict["compression_method"] == "g" else " ") + \
               f" -s {self.param_dict['scale']} " \
               f" -O {compressed_path} {original_path} "

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" -O {reconstructed_path} {compressed_path} "



#!/usr/bin/env python3
"""Tools to define Templates.

Templates are very similar to plugins, but use jinja to transform `.enbt` template files upon installation.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import sys
import argparse
import inspect
import os
import glob
import shutil
import tempfile
import jinja2
import stat

from .installable import Installable, InstallableMeta
import enb.config
from enb.config import options


class MetaTemplate(InstallableMeta):
    def __init__(cls, *args, **kwargs):
        if cls.__name__ != "Template":
            cls.tags.add("template")
        super().__init__(*args, **kwargs)


class Template(Installable, metaclass=MetaTemplate):
    """
    Base class to define templates. Subclasses must be defined in the __plugin__.py file of the template's
    source dir.

    - Templates copy the source dir's contents (except for __plugin__.py) and then transforms
      any `*.enbt` file applying jinja and removing that extension.

    - Templates may require so-called fields in order to produce output.
      These fields can be automatically taken from enb.config.ini (e.g., file-based configuration),
      passed as arguments to the template installation CLI, and programmatically.

    - One or more templates can be installed into an existing directory, the __plugin__.py file is not written
      by default to the installation dir.
    """
    # Map of required field names to their corresponding help
    required_fields_to_help = dict()

    # Files in the template's source dir ending with templatable_extension
    # are subject to jinja templating upon installation.
    templatable_extension = ".enbt"

    @classmethod
    def get_fields(cls, original_fields=None):
        try:
            return cls._fields
        except AttributeError:
            # If there are required fields, satisfy them or fail
            fields = dict(original_fields) if original_fields is not None else dict()
            if cls.required_fields_to_help:
                ini_cli_fields, unused_options = cls.get_field_parser().parse_known_args()
                # Syntax is "plugin install <template> <installation>, so
                # four non-parsed options are expected
                assert len(unused_options) >= 4, (sys.argv, ini_cli_fields, unused_options)
                unused_options = unused_options[4:]
                for field_name in cls.required_fields_to_help:
                    if field_name not in fields:
                        try:
                            fields[field_name] = getattr(ini_cli_fields, field_name)
                            assert fields[field_name] is not None
                        except (KeyError, AssertionError) as ex:
                            raise SyntaxError(
                                f"Missing field {repr(field_name)}. Help for {field_name}:\n"
                                f"{cls.required_fields_to_help[field_name]}\n\n"
                                f"Invoke again with --{field_name}=\"your value\" or with -h for additional help.\n") from ex
                if unused_options:
                    print(f"Warning: unused option{'s' if len(unused_options) > 1 else ''}. \n  - ", end="")
                    print('\n  - '.join(repr(o) for o in unused_options))
                    print(f"NOTE: You can use '' or \"\" to define fields with spaces in them.")
                    print()
            cls._fields = fields
            return fields

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False, fields=None):
        """Install a template into the given dir. See super().install for more information.

        :param installation_dir: directory where the contents of the template are placed.
          It will be created if not existing.
        :param overwrite_destination: if False, a SyntaxError is raised if any of the
          destination contents existed prior to this call. Note that installation_dir
          can already exist, it is the files and directories moved into it that can
          trigger this SyntaxError.
        :param fields: if not None, it must be a dict-like object containing a field to field value
          mapping. If None, it is interpreted as an empty dictionary.
          Required template fields not present in fields will be then read from the CLI
          arguments. If those are not provided, then the default values read from `*.ini`
          configuration files. If any required field cannot not satisfied after this,
          a SyntaxError is raised.
        """
        # If there are required fields, satisfy them or fail
        fields = cls.get_fields(original_fields=fields)

        template_src_dir = os.path.dirname(os.path.abspath(inspect.getfile(cls)))
        for input_path in glob.glob(os.path.join(template_src_dir, "**", "*"), recursive=True):
            if "__pycache__" in input_path:
                continue
            if os.path.basename(input_path) == "__plugin__.py":
                continue

            # By default, the original structure and file names are preserved.
            output_path = os.path.abspath(input_path).replace(
                os.path.abspath(template_src_dir),
                os.path.abspath(installation_dir))

            # Directories are created when found
            if os.path.isdir(input_path):
                os.makedirs(output_path, exist_ok=True)
                continue

            input_is_executable = os.access(input_path, os.X_OK)

            # Files ending in '.enbt' will be identified as templates, processed and stripped of their extension.
            is_templatable = os.path.isfile(input_path) \
                             and os.path.basename(input_path).endswith(cls.templatable_extension)

            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            if is_templatable:
                with tempfile.NamedTemporaryFile(mode="w+") as templated_file:
                    jinja_env = jinja2.Environment(
                        loader=jinja2.FileSystemLoader(os.path.dirname(os.path.abspath(input_path))),
                        autoescape=jinja2.select_autoescape())
                    template = jinja_env.get_template(os.path.basename(input_path))
                    templated_file.write(template.render(**fields))
                    templated_file.flush()
                    templated_file.seek(0)
                    if os.path.exists(output_path[:-len(cls.templatable_extension)]) and not options.force:
                        raise ValueError(
                            f"Error installing template {cls.name}: output file {repr(output_path)} already exists "
                            f"and options.force={options.force}. Run with -f to overwrite.")
                    with open(output_path[:-len(cls.templatable_extension)], "w") as output_file:
                        output_file.write(templated_file.read())
                    if input_is_executable:
                        os.chmod(output_path[:-len(cls.templatable_extension)],
                                 os.stat(output_path[:-len(cls.templatable_extension)]).st_mode | stat.S_IEXEC)
            else:
                if os.path.exists(output_path) and not options.force:
                    raise ValueError(
                        f"Error installing template {cls.name}: output file {repr(output_path)} already exists "
                        f"and options.force={options.force}. Run with -f to overwrite.")
                shutil.copy(input_path, output_path)

        cls.build(installation_dir=installation_dir)
        print(f"Template {repr(cls.name)} successfully installed into {repr(installation_dir)}.")

    @classmethod
    def get_field_parser(cls):
        description = f"Template {repr(cls.name)} installation help."
        if cls.required_fields_to_help:
            description += f"\n\nFields are automatically read from the following paths (in this order):\n"
            description += "\n".join(enb.config.ini.used_config_paths)

            # defined_description = f"\n\nAlready refined fields:"
            defined_field_lines = []
            for field_name in sorted(cls.required_fields_to_help.keys()):
                try:
                    defined_field_lines.append(f"  {field_name} = {enb.config.ini.get_key('template', field_name)}")
                except KeyError:
                    pass
            if defined_field_lines:
                description += f"\n\nFile-defined fields:\n"
                description += "\n".join(defined_field_lines)

        parser = argparse.ArgumentParser(
            prog=f"enb plugin install {cls.name}",
            description=description,
            formatter_class=argparse.RawTextHelpFormatter)
        required_flags_group = parser.add_argument_group(
            "Required flags (use '' or \"\" quoting for fields with spaces)")
        for field_name, field_help in cls.required_fields_to_help.items():
            try:
                default_field_value = enb.config.ini.get_key("template", field_name)
            except KeyError:
                default_field_value = None
            if field_help[-1] != ".":
                field_help += "."
            required_flags_group.add_argument(
                f"--{field_name}",
                default=default_field_value,
                help=field_help,
                metavar=field_name)
        # This argument is for showing help to the user only, since it will have already been parsed
        # by enb.config.ini by the time this is called.
        parser.add_argument(f"--ini", nargs="*", required=False, type=str,
                            help="Additional .ini paths with a [field] section containing field = value lines")
        return parser



#!/usr/bin/env python3
"""Wrappers for the CCSDS 123.0-B-2 codecs.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/01"

import os
import subprocess
import tempfile
import shutil
import sortedcontainers
import math

from enb import icompression
from enb.config import options


class CCSDS_LDC(icompression.LosslessCodec, icompression.WrapperCodec):
    default_large_J = 64

    def __init__(self, bin_dir=None, large_J=None, r=None, restricted_code_options_flag=None,
                 output_header_dir=None):
        """
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        :param large_J: block size (8, 16, 32, or 64), or None for codec's default
        :param r: reference sample period (from 1 to 4096), or None for codec's default
        :param restricted_code_options_flag: use restricted set of code options.
          Must be 0 when n >= 5.
        :param output_header_dir: if not None, LDC headers produced for each
          image are stored in this folder, with a name based on this codec's name
          and the image's full path. Headers stored this way are not reused by the codec
        """
        bin_dir = bin_dir if bin_dir is not None else options.external_bin_base_dir
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        assert os.path.isdir(bin_dir), f"Invalid binary dir {bin_dir}."
        ldc_encoder_path = os.path.join(bin_dir, "ldc_encoder")
        ldc_decoder_path = os.path.join(bin_dir, "ldc_decoder")
        ldc_header_tool_path = os.path.join(bin_dir, "ldc_header_tool")

        param_dict = sortedcontainers.SortedDict()
        large_J = large_J if large_J is not None else self.default_large_J
        assert large_J in (8, 16, 32, 64), f"Invalid large_J value {large_J}"
        param_dict["large_j"] = large_J
        if r is not None:
            assert 1 <= r <= 4096, f"Invalid r value {r}"
            param_dict["r"] = r
        if restricted_code_options_flag is not None:
            param_dict["restricted_code_options_flag"] = restricted_code_options_flag
        for p in (ldc_encoder_path, ldc_decoder_path, ldc_header_tool_path):
            assert os.path.isfile(p), f"Binary {p} does not exist"
            assert os.access(p, os.EX_OK), f"Binary {p} is not executable"
        icompression.WrapperCodec.__init__(
            self, compressor_path=ldc_encoder_path,
            decompressor_path=ldc_decoder_path, param_dict=param_dict)
        self.ldc_header_tool_path = ldc_header_tool_path
        self.output_header_dir = output_header_dir
        
    @property
    def label(self):
        return f"CCSDS 121.0-B-3 J{self.param_dict['large_j']}"

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        with tempfile.NamedTemporaryFile(
                prefix=f"corrected_length_{os.path.basename(original_path)}",
                dir=options.base_tmp_dir) as corrected_length_file, \
                tempfile.NamedTemporaryFile(
                    prefix=f"header_{os.path.basename(original_path)}_") as image_dependent_header_file:
            # Complete full large_j blocks so that LDC does not fail
            shutil.copyfile(original_path, corrected_length_file.name)
            if original_file_info["samples"] % self.param_dict["large_j"] != 0:
                with open(corrected_length_file.name, "ab") as flf:
                    missing_samples = (self.param_dict["large_j"]
                                       - (original_file_info["samples"] % self.param_dict["large_j"]))
                    flf.write(bytes(0 for _ in range(
                        missing_samples * original_file_info["bytes_per_sample"])))

            # Create an image-specific compression header
            header_invocation = \
                f"{self.ldc_header_tool_path} " \
                f"-x n={8 * original_file_info['bytes_per_sample']} " \
                f"-x large_n={os.path.getsize(corrected_length_file.name) // original_file_info['bytes_per_sample']} " \
                f"-x data_sense={0 if original_file_info['signed'] else 1} "
            header_invocation += " ".join(f"-x {k}={v}" for k, v in self.param_dict.items())
            header_invocation += f" {image_dependent_header_file.name}"
            status, output = subprocess.getstatusoutput(header_invocation)
            if status != 0:
                raise Exception("Status = {} != 0.\nInput=[{}].\nOutput=[{}]".format(
                    status, header_invocation, output))
            original_file_info = original_file_info.copy()
            original_file_info["_header_path"] = image_dependent_header_file.name

            # Perform actual compression
            super().compress(original_path=corrected_length_file.name,
                             compressed_path=compressed_path,
                             original_file_info=original_file_info)

            if self.output_header_dir is not None:
                header_output_name = "header_" + self.name + os.path.abspath(os.path.realpath(original_path)).replace(
                    os.sep, "_")
                os.makedirs(self.output_header_dir, exist_ok=True)
                shutil.copyfile(image_dependent_header_file.name,
                                os.path.join(self.output_header_dir, header_output_name))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_file_info['_header_path']} " \
               f"{file_info_to_format_string(original_file_info)} " \
               f"{original_path} {compressed_path}"

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        super().decompress(compressed_path=compressed_path,
                           reconstructed_path=reconstructed_path,
                           original_file_info=original_file_info)
        with open(reconstructed_path, "r+") as rf:
            rf.truncate(original_file_info["samples"]
                        * original_file_info["bytes_per_sample"])

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} " \
               f"{file_info_to_format_string(original_file_info)} " \
               f"{reconstructed_path}"


class CCSDS_LCNL(icompression.NearLosslessCodec, icompression.WrapperCodec):
    ENTROPY_SAMPLE_ADAPTIVE, ENTROPY_HYBRID, ENTROPY_BLOCK_ADAPTIVE = range(3)
    default_entropy_codec_type = ENTROPY_HYBRID

    SAMPLE_ORDER_BIQ, SAMPLE_ORDER_BSQ = range(2)
    default_sample_encoding_order = SAMPLE_ORDER_BSQ  # BSQ

    default_large_P = 3
    LOCAL_SUM_WIDE_NEI, LOCAL_SUM_NARROW_NEI, LOCAL_SUM_WIDE_COL, LOCAL_SUM_NARROW_COL = \
        range(4)
    default_local_sum_type = LOCAL_SUM_WIDE_NEI
    PREDICTION_FULL, PREDICTION_REDUCED = range(2)
    default_prediction_mode = PREDICTION_FULL
    FIDELITY_LOSSLESS, FIDELITY_ABSOLUTE, FIDELITY_RELATIVE, FIDELITY_BOTH = range(4)

    default_large_j = 64
    default_r = 4096
    default_large_r = 64

    def __init__(self, bin_dir=None, absolute_error_limit=None, relative_error_limit=None,
                 entropy_coder_type=None, sample_encoding_order=None, large_p=None,
                 local_sum_type=None, prediction_mode=None, large_j=None, r=None,
                 output_header_dir=None):
        # Process params
        param_dict = sortedcontainers.SortedDict()
        # 
        param_dict["a_vector"] = absolute_error_limit
        param_dict["r_vector"] = relative_error_limit
        # entropy coder type always present
        entropy_coder_type = entropy_coder_type if entropy_coder_type is not None \
            else self.default_entropy_codec_type
        assert entropy_coder_type in (self.ENTROPY_SAMPLE_ADAPTIVE, self.ENTROPY_HYBRID, self.ENTROPY_BLOCK_ADAPTIVE)
        if entropy_coder_type == self.ENTROPY_BLOCK_ADAPTIVE:
            large_j = large_j if large_j is not None else self.default_large_j
            r = r if r is not None else self.default_r
            param_dict["large_j"] = large_j
            param_dict["r"] = r
        param_dict["entropy_coder_type"] = entropy_coder_type
        # Sample encoding order default is BIQ, which is not how images are stored.
        # Therefore, it should always be store in param_dict
        sample_encoding_order = sample_encoding_order if sample_encoding_order is not None \
            else self.default_sample_encoding_order
        assert sample_encoding_order in (self.SAMPLE_ORDER_BIQ, self.SAMPLE_ORDER_BSQ)
        param_dict["sample_encoding_order"] = sample_encoding_order
        #
        large_p = large_p if large_p is not None else self.default_large_P
        assert 0 <= large_p <= 15
        param_dict["large_p"] = large_p
        #
        local_sum_type = local_sum_type if local_sum_type is not None \
            else self.default_local_sum_type
        assert local_sum_type in (self.LOCAL_SUM_WIDE_NEI, self.LOCAL_SUM_NARROW_NEI,
                                  self.LOCAL_SUM_WIDE_COL, self.LOCAL_SUM_NARROW_COL), local_sum_type
        if local_sum_type != self.default_local_sum_type:
            param_dict["local_sum_type"] = local_sum_type
        #
        prediction_mode = prediction_mode if prediction_mode is not None else self.default_prediction_mode
        if prediction_mode is not None and prediction_mode != self.default_prediction_mode:
            assert prediction_mode in (self.PREDICTION_FULL, self.PREDICTION_REDUCED)
            param_dict["prediction_mode"] = prediction_mode

        # Set binary paths
        bin_dir = bin_dir if bin_dir is not None else options.external_bin_base_dir
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        assert os.path.isdir(bin_dir), f"Invalid binary dir {bin_dir}."
        lcnl_encoder_path = os.path.join(bin_dir, "lcnl_encoder")
        lcnl_decoder_path = os.path.join(bin_dir, "lcnl_decoder")
        self.lcnl_header_tool_path = os.path.join(bin_dir, "lcnl_header_tool")
        for p in lcnl_encoder_path, lcnl_decoder_path, self.lcnl_header_tool_path:
            assert os.path.isfile(p), f"Binary file {p} does not exist"
            assert os.access(p, os.EX_OK), f"Binary file {p} is not executable"

        # Complete initialization
        icompression.WrapperCodec.__init__(
            self, compressor_path=lcnl_encoder_path, decompressor_path=lcnl_decoder_path,
            param_dict=param_dict)
        self.output_header_dir = output_header_dir

    def get_image_params(self, original_path, original_file_info):
        """Get the dictionary of parameters to use in the -x k=v options of lcnl_header_tool.
        """
        if self.param_dict['a_vector'] is None and self.param_dict['r_vector'] is None:
            quantizer_fidelity_control_method = self.FIDELITY_LOSSLESS
        elif self.param_dict['a_vector'] is not None and self.param_dict['r_vector'] is None:
            quantizer_fidelity_control_method = self.FIDELITY_ABSOLUTE
        elif self.param_dict['a_vector'] is None and self.param_dict['r_vector'] is not None:
            quantizer_fidelity_control_method = self.FIDELITY_RELATIVE
        else:
            quantizer_fidelity_control_method = self.FIDELITY_BOTH

        image_params = dict(self.param_dict)
        if image_params["a_vector"] is None:
            del image_params["a_vector"]
        else:
            image_params["large_d_large_a"] = math.floor(math.log2(int(image_params["a_vector"]))) + 1 \
                if image_params["a_vector"] > 0 else 1
        if image_params["r_vector"] is None:
            del image_params["r_vector"]
        else:
            image_params["large_d_large_r"] = math.floor(math.log2(int(image_params["r_vector"]))) + 1 \
                if image_params["r_vector"] > 0 else 1

        if original_file_info["bytes_per_sample"] == 1:
            large_d = min(8, original_file_info["dynamic_range_bits"] + 1)
        elif original_file_info["bytes_per_sample"] == 2:
            large_d = max(9, min(16, original_file_info["dynamic_range_bits"] + 1))
        elif original_file_info["bytes_per_sample"] == 4:
            large_d = max(24, min(32, original_file_info["dynamic_range_bits"] + 1))
        else:
            raise ValueError(f"Bytes per sample = {original_file_info['bytes_per_sample']} not supported")

        image_params.update(
            large_n_x=original_file_info["width"],
            large_n_y=original_file_info["height"],
            large_n_z=original_file_info["component_count"],
            sample_type=1 if original_file_info["signed"] else 0,
            large_d=large_d,
            quantizer_fidelity_control_method=quantizer_fidelity_control_method,
            absolute_error_limit_assignment_method=0,  # band independent
            relative_error_limit_assignment_method=0,  # band independent
        )

        return image_params

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        with tempfile.NamedTemporaryFile(
                prefix=f"header_{os.path.basename(original_path)}_") as image_dependent_header_file:
            image_params = self.get_image_params(original_path=original_path, original_file_info=original_file_info)
            header_invocation = f"{self.lcnl_header_tool_path} " + \
                                " ".join(f"-x {k}={v}" for k, v in image_params.items()) + \
                                f" {image_dependent_header_file.name}"
            status, output = subprocess.getstatusoutput(header_invocation)

            if status != 0:
                raise Exception("Status = {} != 0.\nInput=[{}].\nOutput=[{}]".format(
                    status, header_invocation, output))
            original_file_info = original_file_info.copy()
            original_file_info["_header_path"] = image_dependent_header_file.name

            super().compress(original_path=original_path,
                             compressed_path=compressed_path,
                             original_file_info=original_file_info)
            

            if self.output_header_dir is not None:
                header_output_name = "header_" + self.name + os.path.abspath(os.path.realpath(original_path)).replace(
                    os.sep, "_")
                os.makedirs(self.output_header_dir, exist_ok=True)
                shutil.copyfile(image_dependent_header_file.name,
                                os.path.join(self.output_header_dir, header_output_name))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_file_info['_header_path']} " \
               + f"{file_info_to_format_string(original_file_info)} " \
               + f"{original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {file_info_to_format_string(original_file_info)} {reconstructed_path}"

    @property
    def label(self):
        try:
            entropy_coder_type = self.param_dict["entropy_coder_type"]
        except KeyError:
            entropy_coder_type = self.default_entropy_codec_type
        if entropy_coder_type == self.ENTROPY_SAMPLE_ADAPTIVE:
            s = "CCSDS 123.0-B-2 Sample Adaptative"
        elif entropy_coder_type == self.ENTROPY_BLOCK_ADAPTIVE:
            s = "CCSDS 123.0-B-2 Block Adaptative"
        elif entropy_coder_type == self.ENTROPY_HYBRID:
            s = "CCSDS 123.0-B-2 Hybrid"
        else:
            raise ValueError(f"Unexpected entropy coding type {entropy_coder_type}")

        s += f" $a_z={self.param_dict['a_vector']}$" \
            if self.param_dict['a_vector'] is not None else ''
        s += f" $r_z={self.param_dict['r_vector']}$" \
            if self.param_dict['r_vector'] is not None else ''
        # s += f" lossless" \
        #     if self.param_dict['r_vector'] is None \
        #        and self.param_dict['a_vector'] is None else ''
        return s


class CCSDS_LCNL_GreenBook(CCSDS_LCNL):
    """Wrapper for CCSDS_LCNL that uses the default params specified in
    the Green Book.
    """
    # Green book default constants
    default_large_omega = 19
    default_v_min = -1
    default_log_two_t_inc = 6
    default_sample_representative_flag = 1
    default_large_theta = 3
    default_phi_vector = "7"
    default_large_u_max = 18
    default_gamma_star = 6
    default_gamma_0 = 1

    def __init__(self,
                 bin_dir=None,
                 absolute_error_limit=None,
                 relative_error_limit=None,
                 entropy_coder_type=None,
                 output_header_dir=None):
        super().__init__(bin_dir=bin_dir,
                         absolute_error_limit=absolute_error_limit,
                         relative_error_limit=relative_error_limit,
                         entropy_coder_type=entropy_coder_type,
                         output_header_dir=output_header_dir)

    def get_image_params(self, original_path, original_file_info):
        image_params = super().get_image_params(
            original_path=original_path, original_file_info=original_file_info)
        # Add default params if not already present
        for var_name, default_value in [("large_r", self.default_large_r),
                                        ("large_omega", self.default_large_omega),
                                        ("v_min", self.default_v_min),
                                        ("log_two_t_inc", self.default_log_two_t_inc),
                                        ("large_theta", self.default_large_theta),
                                        ("sample_representative_flag", self.default_sample_representative_flag),
                                        ("phi_vector", self.default_phi_vector),
                                        ("large_u_max", self.default_large_u_max),
                                        ("gamma_star", self.default_gamma_star),
                                        ("gamma_0", self.default_gamma_0)]:
            if var_name not in image_params or image_params[var_name] is None:
                image_params[var_name] = default_value

        if any(n in os.path.abspath(os.path.realpath(original_path)).replace(os.sep, "/").lower()
               for n in ("/crism/", "/hyperion/", "/m3/", "/modis/",
                         "SFSI_mantar_Rad_rmnoise-s16be-240x140x452.raw".lower())):
            image_params["local_sum_type"] = self.LOCAL_SUM_WIDE_COL
        else:
            image_params["local_sum_type"] = self.LOCAL_SUM_WIDE_NEI

        if any(n in os.path.abspath(os.path.realpath(original_path)).replace(os.sep, "/").lower()
               for n in ("/airs/",
                         "f011020t01p03r05_sc01.uncal-u16be-224x512x614",
                         "t0477f06-raw-u16be-72x1225x406",
                         "/crism/", "/hyperion/", "/iasi/", "/m3/",
                         "MOD01.A2001222.1200day-nuc-u16be-14x2030x1354.raw".lower(),
                         "MOD01.A2001222.1200day-u16be-14x2030x1354.raw".lower(),
                         "MOD01.A2001222.1200night-u16be-17x2030x1354.raw".lower())):
            image_params["prediction_mode"] = self.PREDICTION_REDUCED
        else:
            image_params["prediction_mode"] = self.PREDICTION_FULL

        if any(n in os.path.abspath(os.path.realpath(original_path)).replace(os.sep, "/").lower()
               for n in ("f060925t01p00r12_sc00.c.img-s16be-224x512x677.raw".lower(),
                         "/sfsi/")):
            image_params["phi_vector"] = 5
        elif any(n in os.path.abspath(os.path.realpath(original_path)).replace(os.sep, "/").lower()
                 for n in ("/airs/", "/aviris/", "/casi/", "/crism/", "/hyperion/",
                           "M3targetB-nuc-u16be-260x512x640.raw".lower(),
                           "M3targetB-u16be-260x512x640.raw".lower())):
            image_params["phi_vector"] = 3
        else:
            image_params["phi_vector"] = 0
        image_params["v_max"] = 7 - image_params["phi_vector"]

        # Use user-defined parameters if present
        for param in ("phi_vector", "psi_vector", "large_theta", "v_max"):
            try:
                image_params[param] = self.param_dict[param]
            except KeyError:
                pass

        return image_params


class CCSDS_LCNL_AdjustedGreenBook(CCSDS_LCNL_GreenBook):
    def get_image_params(self, original_path, original_file_info):
        if any(name in original_path.lower() for name in ("/modis/", "/spot5/", "/hyperion/", "/iasi/")):
            return CCSDS_LCNL.get_image_params(self, original_path=original_path, original_file_info=original_file_info)
        else:
            return super().get_image_params(original_path=original_path, original_file_info=original_file_info)


def file_info_to_format_string(file_info):
    format_string = "s" if file_info["signed"] else "u"
    format_string += str(8 * file_info["bytes_per_sample"])
    format_string += "be" if file_info["big_endian"] else "le"
    return format_string



#!/usr/bin/env python3
"""Abstraction layer to provide parallel processing both locally and on ray clusters.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2022/01/02"

import functools
import os
import time
import datetime
import pathos

from . import config
from .config import options
from . import log
from .log import logger
from . import parallel_ray


def init():
    """If ray is present, this method initializes it.
    If the fallback engine is used, it is ensured that all globals
    are correctly shared with the pool.
    """
    if parallel_ray.is_ray_enabled():
        parallel_ray.init_ray()
    else:
        fallback_init()

def fallback_init():
    """Initialization of the fallback engine. This needs to be called before
    each parallelization, or globals used in the pool might be updated.
    """
    if FallbackFuture.pathos_pool is not None:
        FallbackFuture.pathos_pool.clear()
        FallbackFuture.pathos_pool = None

def chdir_project_root():
    """When invoked, it changes the current working dir to the project's root.
    """
    if parallel_ray.is_parallel_process() and parallel_ray.is_remote_node():
        os.chdir(os.path.expanduser(parallel_ray.RemoteNode.remote_project_mount_path))
    else:
        os.chdir(options.project_root)

def parallel(*args, **kwargs):
    """Decorator for methods intended to run in parallel.

    When ray is available, the .remote() call is performed on the ray decorated function.
    When it is not, a fallback parallelization method is used.

    To run a parallel method `f`, call `f.start` with the arguments you want to pass to f.
    An id object is returned immediately. The result can then be retrieved by calling
    `enb.parallel.get` with the id object.

    Important: parallel calls should not generally read or modify global variables.
    The main exception is enb.config.options, which can be read from parallel calls.
    """
    if parallel_ray.is_ray_enabled():
        return parallel_ray.parallel_decorator(*args, **kwargs)
    else:
        return fallback_parallel_decorator(*args, **kwargs)


def get(ids, **kwargs):
    """Get results for the started ids passed as arguments.

    If timeout is part of kwargs, at most those many seconds are waited.
    Otherwise, this is a blocking call.
    """
    if parallel_ray.is_ray_enabled():
        return parallel_ray.get(ids, **kwargs)
    else:
        return fallback_get(ids, **kwargs)


def get_completed_pending_ids(ids, timeout=0):
    """Given a list of ids returned by start calls, return two lists:
    the first one with the input ids that are ready, and the second
    one with the input ids that are not.
    """
    if parallel_ray.is_ray_enabled():
        return parallel_ray.get_completed_pending_ids(ids, timeout=timeout)
    else:
        return fallback_get_completed_pending_ids(ids, timeout=timeout)


class FallbackFuture:
    """The fallback future is invoked when get is called.
    """
    current_id = 0
    pathos_pool = None

    def __init__(self, f, args, kwargs):
        if self.__class__.pathos_pool is None:
            self.__class__.pathos_pool = pathos.pools.ProcessPool(
                nodes=options.ray_cpu_limit if options.ray_cpu_limit and options.ray_cpu_limit > 0
                else None)
        self.f = f
        self.args = args
        self.kwargs = kwargs
        self.current_id = self.__class__.current_id
        self.__class__.current_id += 1
        self.pathos_result = self.pathos_pool.apipe(f, *args, **kwargs)

    def get(self, **kwargs):
        return self.pathos_result.get(**kwargs)

    def ready(self):
        return self.pathos_result.ready()

    def __hash__(self):
        return hash(self.current_id)


def fallback_parallel_decorator(*decorator_args, **decorator_kwargs):
    """Decorator for methods intended to run in parallel in the local machine.
    """

    def wrapper(f):
        f.start = lambda *_args, **_kwargs: FallbackFuture(f=f, args=_args, kwargs=_kwargs)
        return f

    return wrapper


def fallback_get(ids, **kwargs):
    """Fallback get method when ray is not available.
    """
    return [fallback_future.get(**kwargs) for fallback_future in ids]

def fallback_get_completed_pending_ids(ids, timeout=0):
    """Get two lists, one for completed and one for pending fallback ids.
    """
    complete = []
    pending = []
    for fallback_future in ids:
        if fallback_future.ready():
            complete.append(fallback_future)
        else:
            pending.append(fallback_future)

    time_before = time.time()
    while pending and time.time() - time_before < timeout:
        time.sleep(0.1)
        complete = []
        pending = []
        for fallback_future in ids:
            if fallback_future.ready():
                complete.append(fallback_future)
            else:
                pending.append(fallback_future)


    return complete, pending


class ProgressiveGetter:
    """When an instance is created, the computation of the requested list of calls is started
    in parallel the background (unless they are already running).

    The returned instance is an iterable object. Each to next() with this instance will either
    return the instance if any tasks are still running, or raise StopIteration if all are complete.
    Therefore, instances of this class can be used as the right operand of `in` in for loops.

    A main application of this for-loop approach is to periodically run a code snippet (e.g., for logging)
    while the computation is performed in the background. The loop will continue until all tasks are completed.
    One can then call `ray.get(ray_id_list)` and retrieve the obtained results without any expected delay.

    Note that the for-loop body will always be executed at least once, namely after every potentially
    blocking call to :meth:`ray.wait`.
    """

    def __init__(self, id_list, weight_list=None, iteration_period=1):
        """
        Start the background computation of ids returned by start calls of methods decorated with enb.paralell.parallel.
        After this call, the object is ready to receive next() requests.

        :param id_list: list ids whose values are to be returned
        :param weight_list: if not None, a list of the same length as ray_id list, which contains
          nonnegative values that describe the weight of each task. If provided, they should be highly correlated
          with the computation time of each associated task to provide accurate completion time estimations.
        :param iteration_period: a non-negative value that determines the wait period allowed for ray to
          obtain new results when next() is used. When using this instance in a for loop, it determines approximately
          the periodicity with which the loop body will be executed.
        """
        iteration_period = float(iteration_period)
        if iteration_period < 0:
            raise ValueError(f"Invalid iteration period {iteration_period}: it cannot be negative (but it can be zero)")
        self.full_id_list = list(id_list)
        self.weight_list = weight_list if weight_list is not None else [1] * len(self.full_id_list)
        self.id_to_weight = {i: w for i, w in zip(self.full_id_list, self.weight_list)}
        self.iteration_period = iteration_period
        self.pending_ids = list(self.full_id_list)
        self.completed_ids = []
        self.update_finished_tasks(timeout=0)
        self.start_time = time.time_ns()
        self.end_time = None

    def update_finished_tasks(self, timeout=None):
        """Wait for up to timeout seconds or until ray completes computation
        of all pending tasks. Update the list of completed and pending tasks.
        """
        self.completed_ids, self.pending_ids = get_completed_pending_ids(
            self.full_id_list,
            timeout=timeout if timeout is not None else self.iteration_period)

        try:
            if not self.pending_ids and self.end_time is None:
                self.end_time = time.time_ns()
        except AttributeError:
            self.end_time = time.time_ns()

        assert len(self.completed_ids) + len(self.pending_ids) == len(self.full_id_list)

    def report(self):
        """Return a string that represents the current state of this progressive run.
        """
        if self.pending_ids:
            running_nanos = time.time_ns() - self.start_time
        else:
            running_nanos = self.end_time - self.start_time
        seconds = max(0, running_nanos / 1e9)
        seconds, minutes = seconds - 60 * (seconds // 60), seconds // 60
        minutes, hours = int(minutes % 60), int(minutes // 60)

        total_weight = sum(self.id_to_weight.values())
        completed_weight = sum(self.id_to_weight[i] for i in self.completed_ids)

        time_str = f"{hours:02d}h {minutes:02d}min {seconds:02.3f}s"
        percentage_str = f"{100 * (completed_weight / total_weight):0.1f}%"
        now_str = f"(current time: {datetime.datetime.now()})"

        if self.pending_ids:
            return f"Progress report ({percentage_str}): " \
                   f"{len(self.completed_ids)} / {len(self.full_id_list)} completed tasks. " \
                   f"Elapsed time: {time_str} {now_str}."

        else:
            return f"Progress report: completed all {len(self.full_id_list)} tasks in " \
                   f"{time_str} {now_str}."

    def __iter__(self):
        """This instance is itself iterable.
        """
        return self

    def __next__(self):
        """When next(self) is invoked (directly or using for x in self),
        the lists of complete and pending elements are updated.
        If there are no pending tasks, StopIteration is raised.
        """
        self.update_finished_tasks()
        if not self.pending_ids:
            raise StopIteration
        return self



import os
import shutil
import stat
import glob
import enb.plugins
from enb.config import options

class CompressionExperimentTemplate(enb.plugins.Template):
    """Template for lossy data compression experiments.
    """
    name = "lossy-compression"
    author = ["Miguel Hernández-Cabronero"]
    label = "Template for lossy compression experiments"
    tags = {"project", "data compression"}
    tested_on = {"linux"}

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False, fields=None):
        # Save
        super().install(installation_dir=installation_dir,
                        overwrite_destination=overwrite_destination,
                        fields=fields)
        for py_path in glob.glob(os.path.join(installation_dir, "*.py")):
            os.chmod(py_path, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

        # Copy the default enb.ini file into the project folder
        target_enb_ini = os.path.join(installation_dir, "enb.ini")
        if os.path.exists(target_enb_ini) and not options.force:
            enb.logger.warn(f"File {target_enb_ini} already exists. Not overwriting.")
        else:
            shutil.copyfile(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
                                         f"config", "enb.ini"),
                            os.path.join(installation_dir, f"enb.ini"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Logging utilities for `enb`.

It uses only symbols from .misc, but no other module in enb.
"""
__author__ = "Miguel Hernández-Cabronero"
__date__ = "2021/08/13"

import contextlib
import sys
import builtins
import time

from .misc import ExposedProperty
from .misc import Singleton


class LogLevel:
    """Each of the available logging levels is an instance of this class.
    A level represents a named type of message, with a priority comparable
    to other levels.
    """

    def __init__(self, name, priority=0, prefix=None, help=None):
        """
        :param priority: minimum priority level needed to show this level.
        :param name: unique name for the level.
        :param prefix: prefix when printing messages of this level. If None, a default one
          is used based on the name.
        :param help: optional help explaining the purpose of the level.
        """
        self.name = name
        self.priority = priority
        self.label = prefix
        self.help = help
        if prefix is not None:
            self.prefix = prefix
        else:
            self.prefix = f"[{name[0].upper()}] "

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}:{self.priority})"


class Logger(metaclass=Singleton):
    """Message logging and printing hub for `enb`.

    Messages are only shown if their priority is at least as high as the configured minimum.

    The minimum required level name (from "core" down to "debug") can be selected via the CLI and the file-based configuration
    by setting the `selected_log_level` flag/option.

    You can then modify this minimum value programmatically
    by setting `enb.config.options.minimum_priority_level` to a new :class:`LogLevel` instance,
    such as LOG_VERBOSE or any of the other constants defined above.
    """

    def __init__(self):

        # Available logging levels and their intended usage
        self.levels = [
            LogLevel("core", help="Messages always shown, no matter the priority level"),
            LogLevel("error", help="A critical error that prevents from completing the main task"),
            LogLevel("warn", help="Something wrong or bogus happened, but the main task can be completed"),
            LogLevel("message", help="Task-central messages intended to appear in console"),
            LogLevel("verbose", help="Messages for the interested user, e.g., task progress"),
            LogLevel("info", help="Messages for the very interested user/developer, e.g., detailed task progress"),
            LogLevel("debug", help="Messages for debugging purposes, e.g., traces and watches"),
        ]
        # Assign an integer priority level to each defined level (higher: less priority).
        for i, level in enumerate(self.levels):
            level.priority = i
        # Keep easy access to the levels by their name
        self.name_to_level = {level.name: level for level in self.levels}

        # Minimum priority level required to be printed. It starts by the least restrictive option,
        # to allow logging during the configuration load process without introducing circular
        # references.
        self.selected_log_level = sorted(self.name_to_level.values(), key=lambda level: level.priority)[0]

        # Create a few IDE-friendly attributes that point to the predefined levels by name
        self.level_core = self.get_level("core")
        self.level_error = self.get_level("error")
        self.level_warn = self.get_level("warn")
        self.level_message = self.get_level("message")
        self.level_verbose = self.get_level("verbose")
        self.level_info = self.get_level("info")
        self.level_debug = self.get_level("debug")
        self.show_prefixes = False  # This value is changed based on the file and CLI configuration by enb/__init__.py
        # show_prefix_level determines the level required to include prefixes in any shown messages.
        self.show_prefix_level = self.level_info

    def levels_by_priority(self):
        """Return a list of the available levels, sorted from higher to lower priority.
        """
        return sorted(self.name_to_level.values(), key=lambda level: level.priority)

    def log(self, msg, level, end="\n", file=None, flush=True):
        """Conditionally log a message given its level. It only shares "end" with builtins.print as keyword argument.

        :param msg: message to be logged
        :param level: priority level for the message
        :param end: string appended after the message, if it is shown.
        :param file: file where to log the message, or None to automatically select sys.stdout
        :param flush: if True, the output file is flushed after writing.
        """
        file = sys.stdout if file is None else file

        if level.priority <= self.selected_log_level.priority:
            try:
                last_end = self._last_end
                last_level = self._last_level
            except AttributeError:
                last_end = "\n"
                last_level = self.selected_log_level

            forfeit_prefix = (last_level is level and not last_end.endswith("\n")) \
                             or (self.selected_log_level.priority <= self.show_prefix_level.priority)
            split_message_str = "" if last_level is level or last_end.endswith("\n") else "\n"

            output_msg = f"{split_message_str}" \
                         f"{level.prefix if self.show_prefixes and not forfeit_prefix else ''}" \
                         f"{msg}{end}"

            file.write(output_msg)
            if flush:
                file.flush()

            self._last_end = end
            self._last_level = level

    def core(self, msg, **kwargs):
        """A message of "core" level.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_core, **kwargs)

    def error(self, msg, **kwargs):
        """Log an error message.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_error, **kwargs)

    def warn(self, msg, **kwargs):
        """Log a warning message.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_warn, **kwargs)

    def message(self, msg, **kwargs):
        """Log a regular console message.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_message, **kwargs)

    def verbose(self, msg, **kwargs):
        """Log a verbose console message.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_verbose, **kwargs)

    def info(self, msg, **kwargs):
        """Log an extra-informative console message.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_info, **kwargs)

    def debug(self, msg, **kwargs):
        """Log a debug trace.

        :param kwargs: optional arguments passed to self.log (must be compatible)
        """
        self.log(msg=msg, level=self.level_debug, **kwargs)

    @contextlib.contextmanager
    def log_context(self, msg, level, sep="...", msg_after=None, show_duration=True):
        """Log a message before executing the `with` block code,
        run the block, and log another message when the block is completed.
        The message given the selected priority level, and is only displayed based on `self.selected_log_level`.
        The block of code is executed regardless of the logging options.
        
        :param msg: Message typically describing the
        :param level: Priority level for the shown messages.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow
          single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion. If none, one is automatically selected
          based on msg.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        # Show entry message
        self.log(msg=msg, end=sep, level=level)
        time_before = time.time()

        # Run block
        yield None
        run_time = time.time() - time_before

        # Show exit message
        if msg_after is None:
            space = " "
            try:
                space = space if not self._last_end.endswith("\n") else ""
                msg_after = f"{space}done" if self._last_level is level and self._last_end == sep else f"done ({msg})"
            except AttributeError:
                msg_after = f"{space}done"
        if show_duration:
            msg_after += f" (took {run_time:.2f}s)"
        msg_after += "." if msg_after[-1] != "." else ""
        self.log(msg=msg_after, level=level)

    def core_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of core priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow single-line reporting).
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_core,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def info_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of info priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (\n is not required in it to allow
          single-line reporting.
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_info,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def message_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of message priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow
          single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_message,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def verbose_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of verbose priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow
          single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_verbose,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def info_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of info priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow
          single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_info,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def debug_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of debug priority.
        
        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (newline is not required in it to allow
          single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_debug,
                                sep=sep, msg_after=msg_after, show_duration=show_duration)

    def level_active(self, name, **kwargs):
        """Return True if and only if the given name corresponds to a level with
        priority sufficient given self.min_priority_level.
        """
        return self.name_to_level[name].priority <= self.selected_log_level.priority

    @property
    def core_active(self):
        """Return True if and only if the core level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the core level.
        """
        return self.level_active("core")

    @property
    def error_active(self):
        """Return True if and only if the error level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the error level.
        """
        return self.level_active("error")

    @property
    def warn_active(self):
        """Return True if and only if the warn level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the warn level.
        """
        return self.level_active("warn")

    @property
    def message_active(self):
        """Return True if and only if the message level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the message level.
        """
        return self.level_active("message")

    @property
    def verbose_active(self):
        """Return True if and only if the verbose level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the verbose level.
        """
        return self.level_active("verbose")

    @property
    def info_active(self):
        """Return True if and only if the info level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the info level.
        """
        return self.level_active("info")

    @property
    def debug_active(self):
        """Return True if and only if the debug level is currently active, i.e.,
        the current `self.min_priority_level` has a greater or equal priority value
        than the debug level.
        """
        return self.level_active("debug")

    def report_level_status(self):
        """:return: a string reporting the present logging levels and whether or not they are active.
        """
        lines = [f"{'level':8s}  {'priority':8s}  {'active':6s}"]
        lines.append("-" * len(lines[0]))
        lines.extend(f"{name:8s}  {str(level.priority):8s}  {self.level_active(name)}"
                     for name, level in self.name_to_level.items())
        return "\n".join(lines)

    def get_level(self, name, lower_priority=0):
        """If lower_priority is 0, return the logging level associated with the name passed as argument.
        Otherwise, the aforementioned level's priority is lowered by that numeric amount (positive values means
        less prioritary levels can be selected).

        After that, the available level with the closest priority is chosen.
        """
        # Obtain the logger for the given name
        base_level = self.name_to_level[name]

        # Shift priority if requested - always return something in the list of available levels
        if lower_priority != 0:
            new_priority = base_level.priority + lower_priority
            levels_by_priority = logger.levels_by_priority()
            base_level = levels_by_priority[0]
            for level in levels_by_priority[1:]:
                if level.priority <= new_priority:
                    base_level = level
                else:
                    break

        return base_level

    def replace_print(self, replace=True):
        """When invoked with replace set to True,
        it substitutes the builtin print for a wrapper function that logs the contents
        with "message" priority. If invoked more than once in a row with replace set to True, the second
        and following calls just return the original print function.

        When invoked with replace set to False, the original print function is restored if it is not currently
        substituted with enb's logging method. The original print function is also returned in this case.

        Note that the wrapper function does not admit the file or flush parameters.

        :return: the original builtin's print.
        """
        try:
            _ = self._original_print
            if replace:
                builtins.print = self._original_print
        except AttributeError:
            if not replace:
                # No substitution has been made, builtins contains the original function
                return builtins.print
            else:
                # _original_print is only set to the original print function
                self._original_print = builtins.print
                builtins.print = self.print_to_log

        return self._original_print

    def print_to_log(self, *args, sep=" ", end="\n", file=None, flush=False):
        """Method used to substitute print if configured to do so.
        If file is None, then sys.stdout is used by default.
        """
        self.message(f"{sep.join((str(a) for a in args))}", end=end, file=file, flush=flush)

    def __repr__(self):
        return f"{self.__class__.__name__}(selected={self.selected_log_level})"


# Singleton instance of the logger, shared across modules even if reinstantiated.
logger = Logger()
assert logger is Logger(), "Singleton not working for log.py"

# Expose logging functions
get_level = logger.get_level
log = logger.log
core = logger.core
error = logger.error
warn = logger.warn
message = logger.message
verbose = logger.verbose
info = logger.info
debug = logger.debug

# Expose functions to check whether a level is active or not
core_active = ExposedProperty(instance=logger, property_name="core_active")
error_active = ExposedProperty(instance=logger, property_name="error_active")
warn_active = ExposedProperty(instance=logger, property_name="warn_active")
message_active = ExposedProperty(instance=logger, property_name="message_active")
verbose_active = ExposedProperty(instance=logger, property_name="verbose_active")
info_active = ExposedProperty(instance=logger, property_name="info_active")
debug_active = ExposedProperty(instance=logger, property_name="debug_active")

# Expose report functions

report_level_status = logger.report_level_status



import enb.plugins


class SPDPPlugin(enb.plugins.PluginMake):
    name = "spdp"
    label = "Wrapper for the spdp codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Martin Burtscher"]
    contrib_reference_urls = ["https://userweb.cs.txstate.edu/~burtscher/research/SPDPcompressor/"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Module to handle pgm images
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/08"

import sys
import numpy as np
import re

import enb.isets


def read_pgm(input_path, byteorder='>'):
    """Return image data from a raw PGM file as numpy array.
    Format specification: http://netpbm.sourceforge.net/doc/pgm.html

    (From answer: https://stackoverflow.com/questions/7368739/numpy-and-16-bit-pgm)
    """
    with open(input_path, 'rb') as input_file:
        buffer = input_file.read()
    try:
        header, width, height, maxval = re.search(
            b"(^P5\s(?:\s*#.*[\r\n])*"
            b"(\d+)\s(?:\s*#.*[\r\n])*"
            b"(\d+)\s(?:\s*#.*[\r\n])*"
            b"(\d+)\s(?:\s*#.*[\r\n]\s)*)", buffer).groups()
    except AttributeError as ex:
        raise ex
        raise ValueError(f"Not a raw PGM file: '{input_path}'")
    return np.frombuffer(
        buffer,
        dtype='u1' if int(maxval) < 256 else byteorder + 'u2',
        count=int(width) * int(height),
        offset=len(header)).reshape((int(height), int(width))).swapaxes(0, 1)


def write_pgm(array, bytes_per_sample, output_path, byteorder=">"):
    assert bytes_per_sample in [1, 2], f"bytes_per_sample={bytes_per_sample} not supported"
    assert len(array.shape) == 2, f"Only 2D arrays can be output as PGM"
    assert (array.astype(
        np.int) - array < 2 * sys.float_info.epsilon).all(), f"Only integer values can be stored in PGM"
    assert array.min() >= 0, f"Only positive values can be stored in PGM"
    assert array.max() <= 2 ** (8 * bytes_per_sample) - 1, \
        f"All values should be representable in {bytes_per_sample} bytes " \
        f"(max is {array.max()}, bytes_per_sample={bytes_per_sample})"
    width, height = array.shape
    with open(output_path, "wb") as output_file:
        output_file.write(f"P5\n{width}\n{height}\n{(2 ** (8 * bytes_per_sample)) - 1}\n".encode("utf-8"))
        array.swapaxes(0, 1).astype(f"{byteorder}u{bytes_per_sample}").tofile(output_file)


def pgm_to_raw(input_path, output_path):
    enb.isets.dump_array_bsq(array=read_pgm(input_path), file_or_path=output_path)



import enb.plugins


class FSEPlugin(enb.plugins.PluginMake):
    name = "fse"
    label = "Wrappers for FSE codecs"
    tags = {"data compression", "codec"}
    contrib_authors = ["Yann Collet"]
    contrib_reference_urls = ["https://github.com/Cyan4973/FiniteStateEntropy"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/fse_git.zip?raw=true",
         "fse_git.zip")]
    tested_on = {"linux"}



#!/usr/bin/env python3

import os
import glob
import pickle
import unittest
import string
import numpy as np
import pathos

import enb.atable
import enb.atable as atable


class Subclass(atable.ATable):
    @atable.column_function("index_length")
    def set_index_length(self, index, row):
        row["index_length"] = len(index)


class DefinitionModesTable(atable.ATable):
    """Example ATable subclass that exemplifies multiple ways of defining columns.
    """

    def column_a(self, index, row):
        """Methods that start with column_* are automatically recognized as column definitions.
        The returned value is the value set into the appropriate column of row.
        """
        return 1

    def column_b(self, index, row):
        """Previously defined column values for this row can be used in other columns.
        Columns are computed in the order they are defined."""
        return row["a"] + 1

    @atable.column_function("c")
    def set_column_c(self, index, row):
        """The @enb.atable.column_function decorator can be used to explicitly mark class methods as
        column functions, i.e., functions that set one or more column functions.

        Functions decorated with @enb.atable.column_function must explicitly edit the row parameter
        to update the column function being set.
        """
        row["c"] = row["b"] + 1

    @atable.column_function("d")
    def set_column_d(self, index, row):
        """The _column_name is automatically be defined before calling a decorated (or column_*) function.
        This way, the function code needs not change if the column name is renamed.
        """
        row[_column_name] = row["c"] + 1

    def ignore_column_f(self, index, row):
        """Any number of methods can be defined normally in the ATable subclass.
        These are not invoked automatically by enb.
        """
        raise Exception("This method is never invoked")

    @atable.column_function(
        [atable.ColumnProperties("e"),
         atable.ColumnProperties("f", label="label for f", plot_min=0, plot_max=10)]
    )
    def set_multiple_colums(self, index, row):
        """Multiple columns can be set with the same decorated column function
        (not with undecorated column_* methods).

        The @enb.atable.column_function decorator accepts a list of atable.ColumnProperties instances,
        one per column being set by this function.

        Check out the documentation for atable.ColumnProperties, as it allows providing hints
        for plotting any column individually.
        See https://miguelinux314.github.io/experiment-notebook/using_analyzer_subclasses.html
        for more details.
        """
        row["e"] = row["d"] + 1
        row["f"] = row["e"] + 1

    @atable.column_function(
        atable.ColumnProperties("constant_one", plot_min=1, plot_max=1),
        "constant_zero",
        atable.ColumnProperties("first_ten_numbers", has_iterable_values=True),
        atable.ColumnProperties("ascii_table", has_dict_values=True),
    )
    def set_mixed_type_columns(self, index, row):
        row["constant_one"] = 1
        row["constant_zero"] = 0
        row["first_ten_numbers"] = list(range(10))
        row["ascii_table"] = {l: ord(l) for l in string.ascii_letters}


class TestATable(unittest.TestCase):
    def test_subclassing(self):
        """Verify that ATable can be subclassed with well defined behavior
        """
        df_length = 5
        sc = Subclass(index="index")
        assert sc.column_to_properties["index_length"].fun.__qualname__ == Subclass.set_index_length.__qualname__, \
            (sc.column_to_properties["index_length"].fun, Subclass.set_index_length,
             Subclass.set_index_length.__qualname__)
        df = sc.get_df(target_indices=["a" * i for i in range(df_length)])
        assert len(df) == df_length
        assert (df["index_length"].values == range(df_length)).all(), \
            (df["index_length"].values)

    def test_column_definition_modes(self):
        """Test the different methods of adding columns to a table and verify that they work properly.

        Sets the columns from a to f with values from 1 to 6.
        """

        table = DefinitionModesTable()
        df = table.get_df(target_indices=["val0"])
        assert len(df) == 1, len(df)
        row = df.iloc[0]
        for i, c in enumerate((chr(o) for o in range(ord("a"), ord("f") + 1)), start=1):
            assert row[c] == i

        assert row["constant_zero"] == 0, row["constant_zero"]
        assert row["constant_one"] == 1, row["constant_one"]
        assert row["first_ten_numbers"] == list(range(10)), row["first_ten_numbers"]
        assert len(row["ascii_table"]) == len(string.ascii_letters), len(row["ascii_table"])
        assert all(v == ord(k) for k, v in row["ascii_table"].items()), row["ascii_table"]


class TestFailingTable(unittest.TestCase):
    def test_always_failing_column(self):
        ft = FailingTable()
        target_indices = string.ascii_letters
        original_verbose = enb.config.options.verbose

        try:
            enb.config.options.verbose = -1
            ft.get_df(target_indices=target_indices)
            enb.config.options.verbose = original_verbose
            raise RuntimeError("The previous call should have failed")
        except enb.atable.ColumnFailedError as ex:
            assert len(ex.exception_list) == len(target_indices)
            pass
        except (NotImplementedError, pickle.PickleError):
            pass
        finally:
            enb.config.options.verbose = original_verbose


class FailingTable(enb.atable.ATable):
    def column_failing(self, index, row):
        raise NotImplementedError("I am expected to crash - no worries!")


class TestSummaryTable(unittest.TestCase):
    def test_summary_table(self):
        base_table = enb.sets.FilePropertiesTable()

        target_paths = [
            p for p in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "*"))
            if os.path.isfile(p)]

        base_df = base_table.get_df(target_indices=target_paths)

        summary_table = enb.atable.SummaryTable(full_df=base_df)
        summary_df = summary_table.get_df()

        assert summary_df.iloc[0]["group_label"].lower() == "all", summary_df.iloc[0]["group_label"]
        assert summary_df.iloc[0]["group_size"] == len(target_paths)


class CustomType:
    def __init__(self, custom_prop):
        self.custom_prop = custom_prop
        self.other_prop = 2 * custom_prop


class TestObjectColumns(unittest.TestCase):
    def test_object_values(self):
        df = TypesTable().get_df(target_indices=string.ascii_letters)
        assert np.all(df["custom_type_column"].apply(type).str.endswith(CustomType.__name__)), \
            ", ".join(str(type(t)) for t in df["custom_type_column"].unique())
        assert np.all(df["custom_type_column"].apply(lambda ct: ct.custom_prop) == 1)
        assert np.all(df["custom_type_column"].apply(lambda ct: ct.other_prop) == 2)


class TypesTable(enb.atable.ATable):
    @enb.atable.column_function(
        "uppercase",
        "lowercase",
        enb.atable.ColumnProperties(
            "first_last_iterable",
            label="First and last characters of the index",
            has_iterable_values=True),
        enb.atable.ColumnProperties(
            "first_last_dict",
            label="First and last characters of the index",
            has_dict_values=True),
        enb.atable.ColumnProperties(
            "custom_type_column", has_object_values=True)
    )
    def set_columns(self, index, row):
        row["uppercase"] = index.upper()
        row["lowercase"] = index.lower()
        row["first_last_iterable"] = (index[0], index[-1]) if index else []
        row["first_last_dict"] = dict(first=index[0], last=index[-1]) if index else {}
        row["custom_type_column"] = CustomType(custom_prop=len(index))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Unit tests for isets.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/07"

import unittest
import tempfile
import glob
import os
import numpy as np
import shutil
import subprocess

import test_all
from enb import isets


class TestiSets(unittest.TestCase):

    def test_array_load_dump(self):
        """Test that loading and dumping of numpy arrays indexed by [x,y,z] is correctly made.
        """
        width = 4
        height = 2
        component_count = 3

        for is_float in [True, False]:
            bytes_per_sample_values = [2, 4, 8] if is_float else [1, 2, 4, 8]
            for bytes_per_sample in bytes_per_sample_values:
                for signed in [True, False]:
                    value_factors = [1] if not signed else [1, -1]
                    for value_factor in value_factors:
                        big_endian_values = [True] if (bytes_per_sample == 1 or is_float) \
                            else [True, False]
                        for big_endian in big_endian_values:
                            row = dict(
                                width=width, height=height, component_count=component_count, float=is_float,
                                signed=signed, bytes_per_sample=bytes_per_sample, big_endian=big_endian)

                            array = np.zeros((width, height, component_count),
                                             dtype=isets.iproperties_row_to_numpy_dtype(row))
                            i = 0
                            for z in range(component_count):
                                for y in range(height):
                                    for x in range(width):
                                        array[x, y, z] = i * value_factor
                                        i += 1

                            _, temp_file_path = tempfile.mkstemp(
                                suffix=f"_{isets.iproperties_row_to_sample_type_tag(row)}.raw")
                            try:
                                dtype_string = isets.iproperties_row_to_numpy_dtype(row)
                                isets.dump_array_bsq(array, temp_file_path, mode="wb",
                                                     dtype=dtype_string)
                                loaded_array = isets.load_array_bsq(file_or_path=temp_file_path,
                                                                    image_properties_row=row)
                                assert (array == loaded_array).all(), (array, loaded_array)
                                assert os.path.getsize(temp_file_path) \
                                       == width * height * component_count * bytes_per_sample
                            finally:
                                try:
                                    os.remove(temp_file_path)
                                except OSError:
                                    pass


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Image compression experiment module
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/01"

import filecmp
import hashlib
import os
import tempfile
import time
import collections
import functools
import shutil
import math
import numpy as np
import imageio
import subprocess

import numpngw
from astropy.io import fits
from scipy import signal
from scipy.ndimage.filters import convolve

import enb
import enb.atable
from enb import atable
from enb import experiment
from enb import isets
from enb import tcall
from enb.atable import indices_to_internal_loc
from enb.config import options


class CompressionResults:
    """Base class that defines the minimal fields that are returned by a
    call to a coder's compress() method (or produced by
    the CompressionExperiment instance).
    """

    def __init__(self, codec_name=None, codec_param_dict=None, original_path=None,
                 compressed_path=None, side_info_files=[], compression_time_seconds=None):
        """
        :param codec_name: codec's reported_name
        :param codec_param_dict: dictionary of parameters to the codec
        :param original_path: path to the input original file
        :param compressed_path: path to the output compressed file
        :param side_info_files: list of file paths with side information
        :param compression_time_seconds: effective average compression time in seconds
        """
        self.codec_name = codec_name
        self.codec_param_dict = codec_param_dict
        self.original_path = original_path
        self.compressed_path = compressed_path
        self.side_info_files = side_info_files
        self.compression_time_seconds = compression_time_seconds


class DecompressionResults:
    """Base class that defines the minimal fields that are returned by a
    call to a coder's decompress() method (or produced by
    the CompressionExperiment instance).
    """

    def __init__(self, codec_name=None, codec_param_dict=None, compressed_path=None,
                 reconstructed_path=None, side_info_files=[], decompression_time_seconds=None):
        """
        :param codec_name: codec's reported_name
        :param codec_param_dict: dictionary of parameters to the codec
        :param compressed_path: path to the output compressed file
        :param reconstructed_path: path to the reconstructed file after decompression
        :param side_info_files: list of file paths with side information
        :param decompression_time_seconds: effective decompression time in seconds
        """
        self.codec_name = codec_name
        self.codec_param_dict = codec_param_dict
        self.compressed_path = compressed_path
        self.reconstructed_path = reconstructed_path
        self.side_info_files = side_info_files
        self.decompression_time_seconds = decompression_time_seconds


class CompressionException(Exception):
    """Base class for exceptions occurred during a compression instance
    """

    def __init__(self, original_path=None, compressed_path=None, file_info=None,
                 status=None, output=None):
        super().__init__(f"status={status}, output={output}")
        self.original_path = original_path
        self.compressed_path = compressed_path
        self.file_info = file_info
        self.status = status
        self.output = output


class DecompressionException(Exception):
    """Base class for exceptions occurred during a decompression instance
    """

    def __init__(self, compressed_path=None, reconstructed_path=None,
                 file_info=None, status=None, output=None):
        super().__init__(f"status={status}, output={output}")
        self.reconstructed_path = reconstructed_path
        self.compressed_path = compressed_path
        self.file_info = file_info
        self.status = status
        self.output = output


class AbstractCodec(experiment.ExperimentTask):
    """Base class for all codecs
    """

    def __init__(self, param_dict=None):
        super().__init__(param_dict=param_dict)

    @property
    def name(self):
        """Name of the codec. Subclasses are expected to yield different values
        when different parameters are used. By default, the class name is folled
        by all elements in self.param_dict sorted alphabetically are included
        in the name."""
        name = f"{self.__class__.__name__}"
        if self.param_dict:
            name += "__" + "_".join(f"{k}={v}" for k, v in sorted(self.param_dict.items()))
        return name

    @property
    def label(self):
        """Label to be displayed for the codec. May not be strictly unique nor fully informative.
        By default, the original name is returned
        """
        return self.name

    @property
    def label_with_params(self):
        return self.label + " " + ", ".join(f"{enb.atable.clean_column_name(k)}"
                                            f"={self.param_dict[k]}"
                                            for k in sorted(self.param_dict.keys()))

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        """Compress original_path into compress_path using param_dict as params.
        :param original_path: path to the original file to be compressed
        :param compressed_path: path to the compressed file to be created
        :param original_file_info: a dict-like object describing original_path's properties (e.g., geometry), or None.
        :return: (optional) a CompressionResults instance, or None (see compression_results_from_paths)
        """
        raise NotImplementedError()

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        """Decompress compressed_path into reconstructed_path using param_dict
        as params (if needed).

        :param compressed_path: path to the input compressed file
        :param reconstructed_path: path to the output reconstructed file
        :param original_file_info: a dict-like object describing original_path's properties
          (e.g., geometry), or None. Should only be actually used in special cases,
          since codecs are expected to store all needed metainformation in the compressed file.
        :return: (optional) a DecompressionResults instance, or None (see decompression_results_from_paths)
        """
        raise NotImplementedError()

    def compression_results_from_paths(self, original_path, compressed_path):
        """Get the default CompressionResults instance corresponding to
        the compression of original_path into compressed_path
        """
        return CompressionResults(
            codec_name=self.name,
            codec_param_dict=self.param_dict,
            original_path=original_path,
            compressed_path=compressed_path,
            side_info_files=[],
            compression_time_seconds=None)

    def decompression_results_from_paths(self, compressed_path, reconstructed_path):
        return DecompressionResults(
            codec_name=self.name,
            codec_param_dict=self.param_dict,
            compressed_path=compressed_path,
            reconstructed_path=reconstructed_path,
            side_info_files=[],
            decompression_time_seconds=None)

    def __repr__(self):
        return f"<{self.__class__.__name__}" \
               f"({', '.join(repr(param) + '=' + repr(value) for param, value in self.param_dict.items())})>"


class LosslessCodec(AbstractCodec):
    """A AbstractCodec that identifies itself as lossless
    """
    pass


class LossyCodec(AbstractCodec):
    """A AbstractCodec that identifies itself as lossy
    """
    pass


class NearLosslessCodec(LossyCodec):
    pass


class WrapperCodec(AbstractCodec):
    """A codec that uses an external process to compress and decompress.
    """

    def __init__(self, compressor_path, decompressor_path, param_dict=None, output_invocation_dir=None,
                 signature_in_name=False):
        """
        :param compressor_path: path to the executable to be used for compression
        :param decompressor_path: path to the executable to be used for decompression
        :param param_dict: name-value mapping of the parameters to be used for compression
        :param output_invocation_dir: if not None, invocation strings are stored in this directory
          with name based on the codec and the image's full path.
        :pram signature_in_name: if True, the default codec name includes part of the hexdigest of the
          compressor and decompressor binaries being used
        """
        super().__init__(param_dict=param_dict)
        self.signature_in_name = False
        if os.path.isfile(compressor_path):
            self.compressor_path = compressor_path
        else:
            self.compressor_path = shutil.which(compressor_path)
        if os.path.exists(decompressor_path):
            self.decompressor_path = decompressor_path
        else:
            self.decompressor_path = shutil.which(decompressor_path)

        if not self.compressor_path or not os.path.isfile(self.compressor_path):
            raise FileNotFoundError(f"Compressor path {repr(compressor_path)} is not available "
                                    f"for {self.__class__.__name__}")
        if not self.decompressor_path or not os.path.isfile(self.decompressor_path):
            raise FileNotFoundError(f"Decompressor path {repr(compressor_path)} is not available "
                                    f"for {self.__class__.__name__}")
        self.output_invocation_dir = output_invocation_dir
        if self.output_invocation_dir is not None:
            os.makedirs(self.output_invocation_dir, exist_ok=True)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        """Return a string (shell style) with the parameters
        to be passed to the compressor.

        Same parameter semantics as :meth:`AbstractCodec.compress`.

        :param original_file_info: a dict-like object describing original_path's properties
          (e.g., geometry), or None
        """
        raise NotImplementedError()

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        """Return a string (shell style) with the parameters
        to be passed to the decompressor.
        Same parameter semantics as :meth:`AbstractCodec.decompress()`.

        :param original_file_info: a dict-like object describing original_path's properties
          (e.g., geometry), or None
        """
        raise NotImplementedError()

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        compression_params = self.get_compression_params(
            original_path=original_path,
            compressed_path=compressed_path,
            original_file_info=original_file_info)
        invocation = f"{self.compressor_path} {compression_params}"
        enb.logger.info(f"[{self.name}] Invocation: '{invocation}'")
        try:
            status, output, measured_time = tcall.get_status_output_time(invocation=invocation)
            enb.logger.debug(
                f"[{self.name}] Compression OK; invocation={invocation} - status={status}; output={output}")
        except tcall.InvocationError as ex:
            raise CompressionException(
                original_path=original_path,
                compressed_path=compressed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        compression_results = self.compression_results_from_paths(
            original_path=original_path, compressed_path=compressed_path)
        compression_results.compression_time_seconds = measured_time

        if self.output_invocation_dir is not None:
            invocation_name = "invocation_compression_" \
                              + self.name \
                              + os.path.abspath(os.path.realpath(original_file_info["file_path"])).replace(os.sep, "_")
            with open(os.path.join(self.output_invocation_dir, invocation_name), "w") as invocation_file:
                invocation_file.write(f"Original path: {original_path}\n"
                                      f"Compressed path: {compressed_path}\n"
                                      f"Codec: {self.name}\n"
                                      f"Invocation: {invocation}\n"
                                      f"Status: {status}\n"
                                      f"Output: {output}\n"
                                      f"Measured time: {measured_time}")

        return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        decompression_params = self.get_decompression_params(
            compressed_path=compressed_path,
            reconstructed_path=reconstructed_path,
            original_file_info=original_file_info)
        invocation = f"{self.decompressor_path} {decompression_params}"
        enb.logger.info(f"WrapperCodec:decompress invocation={invocation}")
        try:
            status, output, measured_time = tcall.get_status_output_time(invocation)
            enb.logger.debug(
                f"[{self.name}] Compression OK; invocation={invocation} - status={status}; output={output}")
        except tcall.InvocationError as ex:
            raise DecompressionException(
                compressed_path=compressed_path,
                reconstructed_path=reconstructed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        decompression_results = self.decompression_results_from_paths(
            compressed_path=compressed_path, reconstructed_path=reconstructed_path)

        decompression_results.decompression_time_seconds = measured_time

        if self.output_invocation_dir is not None:
            invocation_name = "invocation_decompression_" \
                              + self.name \
                              + os.path.abspath(os.path.realpath(
                original_file_info["file_path"] if original_file_info is not None else compressed_path)).replace(os.sep,
                                                                                                                 "_")
            with open(os.path.join(self.output_invocation_dir, invocation_name), "w") as invocation_file:
                invocation_file.write(f"Compressed path: {compressed_path}\n"
                                      f"Reconstructed path: {reconstructed_path}\n"
                                      f"Codec: {self.name}\n"
                                      f"Invocation: {invocation}\n"
                                      f"Status: {status}\n"
                                      f"Output: {output}\n"
                                      f"Measured time: {measured_time}")

        return decompression_results

    @staticmethod
    @functools.lru_cache(maxsize=2)
    def get_binary_signature(binary_path):
        """Return a string with a (hopefully) unique signature for
        the contents of binary_path. By default, the first 5 digits
        of the sha-256 hexdigest are returned.
        """
        hasher = hashlib.sha256()
        with open(binary_path, "rb") as open_file:
            hasher.update(open_file.read())
        return hasher.hexdigest()[:5]

    @property
    def name(self):
        """Return the codec's name and parameters, also including the encoder
        and decoder hash summaries (so that changes in the reference binaries
        can be easily detected)
        """
        signature = None
        if self.signature_in_name:
            compressor_signature = self.get_binary_signature(self.compressor_path)
            decompressor_signature = self.get_binary_signature(self.decompressor_path)
            if compressor_signature and decompressor_signature:
                if compressor_signature == decompressor_signature:
                    signature = f"{compressor_signature}"
                else:
                    signature = f"{compressor_signature}_{compressor_signature}"

        name = f"{self.__class__.__name__}_{signature}" if signature is not None else self.__class__.__name__
        if self.param_dict:
            name += "__" + "_".join(f"{k}={v}" for k, v in sorted(self.param_dict.items()))
        return name


class FITSWrapperCodec(WrapperCodec):
    """Raw images are coded into FITS before compression with the wrapper,
    and FITS is decoded to raw after decompression.
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        img = enb.isets.load_array_bsq(
            file_or_path=original_path, image_properties_row=original_file_info)
        with tempfile.NamedTemporaryFile(suffix=".fits") as tmp_file:
            os.remove(tmp_file.name)
            array = img.swapaxes(0, 2)
            hdu = fits.PrimaryHDU(array)
            hdu.writeto(tmp_file.name)

            if os.path.exists(compressed_path):
                os.remove(compressed_path)
            compression_results = super().compress(original_path=tmp_file.name,
                                                   compressed_path=compressed_path,
                                                   original_file_info=original_file_info)
            cr = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            cr.compression_time_seconds = max(0, compression_results.compression_time_seconds)
            return cr

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):

        with tempfile.NamedTemporaryFile(suffix=".fits") as tmp_file:
            os.remove(tmp_file.name)
            decompression_results = super().decompress(
                compressed_path=compressed_path, reconstructed_path=tmp_file.name)

            hdul = fits.open(tmp_file.name)
            assert len(hdul) == 1
            data = hdul[0].data.transpose()
            header = hdul[0].header

            if header['NAXIS'] == 2:
                if header['BITPIX'] < 0:
                    dtype_name = f'float{-header["BITPIX"]}'
                elif header['BITPIX'] > 0:
                    dtype_name = f'>u{header["BITPIX"] // 8}'
                else:
                    raise ValueError(f"Invalid bitpix {header['BITPIX']}")
            elif header['NAXIS'] == 3:
                if header['BITPIX'] < 0:
                    dtype_name = f'float{-header["BITPIX"]}'
                elif header['BITPIX'] > 0:
                    dtype_name = f'>u{header["BITPIX"] // 8}'
                else:
                    raise ValueError(f"Invalid bitpix {header['BITPIX']}")
            else:
                raise Exception(f"Invalid header['NAXIS'] = {header['NAXIS']}")

            enb.isets.dump_array_bsq(array=data, file_or_path=reconstructed_path, dtype=dtype_name)
            decompression_results.compressed_path = compressed_path
            decompression_results.reconstructed_path = reconstructed_path
            return decompression_results


class PNGWrapperCodec(WrapperCodec):
    """Raw images are coded into PNG before compression with the wrapper,
    and PNG is decoded to raw after decompression.
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        img = enb.isets.load_array_bsq(
            file_or_path=original_path, image_properties_row=original_file_info)

        with tempfile.NamedTemporaryFile(suffix=".png") as tmp_file:
            numpngw.write_png(tmp_file.name, img)
            compression_results = super().compress(
                original_path=tmp_file.name,
                compressed_path=compressed_path,
                original_file_info=original_file_info)
            cr = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            cr.compression_time_seconds = max(
                0, compression_results.compression_time_seconds)
            return cr

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with tempfile.NamedTemporaryFile(suffix=".png") as tmp_file:
            decompression_results = super().decompress(
                compressed_path=compressed_path, reconstructed_path=tmp_file.name)

            invocation = f"file {tmp_file.name}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception("Status = {} != 0.\nInput=[{}].\nOutput=[{}]".format(
                    status, invocation, output))
            img = imageio.imread(tmp_file.name, "png")
            img.swapaxes(0, 1)
            assert len(img.shape) in [2, 3, 4]
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)

            dtype = ">"
            dtype += "i" if original_file_info["signed"] else "u"
            dtype += f"{original_file_info['bytes_per_sample']}"

            enb.isets.dump_array_bsq(img, file_or_path=reconstructed_path, dtype=dtype)

            dr = self.decompression_results_from_paths(
                compressed_path=compressed_path, reconstructed_path=reconstructed_path)
            dr.decompression_time_seconds = decompression_results.decompression_time_seconds


class PGMWrapperCodec(WrapperCodec):
    """Raw images are coded into PNG before compression with the wrapper,
    and PNG is decoded to raw after decompression.
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        assert original_file_info["component_count"] == 1, "PGM only supported for 1-component images"
        assert original_file_info["bytes_per_sample"] in [1, 2], "PGM only supported for 8 or 16 bit images"
        img = enb.isets.load_array_bsq(
            file_or_path=original_path, image_properties_row=original_file_info)

        with tempfile.NamedTemporaryFile(suffix=".pgm", mode="wb") as tmp_file:
            numpngw.imwrite(tmp_file.name, img)
            with open(tmp_file, "rb") as raw_file:
                contents = raw_file.read()
            os.remove(tmp_file)
            with open(tmp_file, "wb") as pgm_file:
                tmp_file.write(bytes(f"P6\n"
                                     f"{original_file_info['width']} {original_file_info['height']}\n"
                                     f"{255 if original_file_info['bytes_per_sample'] == 1 else 65535}\n"))
                tmp_file.write(contents)

            compression_results = super().compress(original_path=tmp_file.name,
                                                   compressed_path=compressed_path,
                                                   original_file_info=original_file_info)
            cr = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            cr.compression_time_seconds = max(
                0, compression_results.compression_time_seconds)
            return cr

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with tempfile.NamedTemporaryFile(suffix=".pgm") as tmp_file:
            decompression_results = super().decompress(
                compressed_path=compressed_path, reconstructed_path=tmp_file.name)
            img = imageio.imread(tmp_file.name, "pgm")
            img.swapaxes(0, 1)
            assert len(img.shape) in [2, 3, 4]
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)
            enb.isets.dump_array_bsq(img, file_or_path=reconstructed_path)

            dr = self.decompression_results_from_paths(
                compressed_path=compressed_path, reconstructed_path=reconstructed_path)
            dr.decompression_time_seconds = decompression_results.decompression_time_seconds


class CompressionExperiment(experiment.Experiment):
    """This class allows seamless execution of compression experiments.

    In the functions decorated with @atable,column_function, the row argument
    contains two magic properties, compression_results and decompression_results.
    These give access to the :class:`CompressionResults` and :class:`DecompressionResults`
    instances resulting respectively from compressing and decompressing
    according to the row index parameters. The paths referenced in the compression
    and decompression results are valid while the row is being processed, and
    are disposed of afterwards.
    Also, the image_info_row attribute gives access to the image metainformation
    (e.g., geometry)
    """
    dataset_files_extension = "raw"
    default_file_properties_table_class = enb.isets.ImagePropertiesTable
    check_lossless = True
    row_wrapper_column_name = "_codec_wrapper"

    class CompressionDecompressionWrapper:
        """This class is instantiated for each row of the table, and added to a temporary
        column row_wrapper_column_name. Column-setting methods can then access this wrapper,
        and in particular its `compression_results` and `decompression_results` properties,
        which will run compression and decompression at most once. This way, many columns can
        be defined independently without needing to compress and decompress for each one.
        """

        def __init__(self, file_path, codec, image_info_row):
            self.file_path = file_path
            self.codec = codec
            self.image_info_row = image_info_row
            self._compression_results = None
            self._decompression_results = None

        @property
        def compression_results(self):
            """Perform the actual compression experiment for the selected row.
            """
            if self._compression_results is None:
                _, tmp_compressed_path = tempfile.mkstemp(
                    dir=options.base_tmp_dir,
                    prefix=f"compressed_{os.path.basename(self.file_path)}_")
                try:
                    measured_times = []

                    enb.logger.info(f"Executing compression {self.codec.name} on {self.file_path} "
                                    f"[{options.repetitions} times]")
                    for repetition_index in range(options.repetitions):
                        enb.logger.info(
                            f"Executing compression {self.codec.name} on {self.file_path} "
                            f"[rep{repetition_index + 1}/{options.repetitions}]")
                        time_before = time.time()
                        self._compression_results = self.codec.compress(original_path=self.file_path,
                                                                        compressed_path=tmp_compressed_path,
                                                                        original_file_info=self.image_info_row)

                        if not os.path.isfile(tmp_compressed_path) \
                                or os.path.getsize(tmp_compressed_path) == 0:
                            raise CompressionException(
                                original_path=self.file_path, compressed_path=tmp_compressed_path,
                                file_info=self.image_info_row,
                                output=f"Compression didn't produce a file (or it was empty) {self.file_path}")

                        wall_compression_time = time.time() - time_before
                        if self._compression_results is None:
                            enb.logger.info(f"[W]arning: codec {self.codec.name} did not report execution times. "
                                            f"Using wall clock instead (might be inaccurate)")
                            self._compression_results = self.codec.compression_results_from_paths(
                                original_path=self.file_path, compressed_path=tmp_compressed_path)
                            self._compression_results.compression_time_seconds = wall_compression_time

                        measured_times.append(self._compression_results.compression_time_seconds)
                        if repetition_index < options.repetitions - 1:
                            os.remove(tmp_compressed_path)

                    self._compression_results.compression_time_seconds = sum(measured_times) / len(measured_times)
                except Exception as ex:
                    os.remove(tmp_compressed_path)
                    raise ex

            return self._compression_results

        @property
        def decompression_results(self):
            """Perform the actual decompression experiment for the selected row.
            """
            if self._decompression_results is None:
                _, tmp_reconstructed_path = tempfile.mkstemp(
                    prefix=f"reconstructed_{os.path.basename(self.file_path)}",
                    dir=options.base_tmp_dir)
                try:
                    measured_times = []
                    with enb.logger.info_context(
                            f"Executing decompression {self.codec.name} on {self.file_path} "
                            f"[{options.repetitions} times]"):
                        for repetition_index in range(options.repetitions):
                            enb.logger.info(f"Executing decompression {self.codec.name} on {self.file_path} "
                                            f"[rep{repetition_index + 1}/{options.repetitions}]")

                            time_before = time.time()
                            self._decompression_results = self.codec.decompress(
                                compressed_path=self.compression_results.compressed_path,
                                reconstructed_path=tmp_reconstructed_path,
                                original_file_info=self.image_info_row)

                            wall_decompression_time = time.time() - time_before
                            if self._decompression_results is None:
                                enb.logger.info(
                                    f"Codec {self.codec.name} did not report execution times. "
                                    f"Using wall clock instead (might be inaccurate)")
                                self._decompression_results = self.codec.decompression_results_from_paths(
                                    compressed_path=self.compression_results.compressed_path,
                                    reconstructed_path=tmp_reconstructed_path)
                                self._decompression_results.decompression_time_seconds = wall_decompression_time

                            if not os.path.isfile(tmp_reconstructed_path) or os.path.getsize(
                                    self._decompression_results.reconstructed_path) == 0:
                                raise CompressionException(
                                    original_path=self.compression_results.original_path,
                                    compressed_path=self.compression_results.compressed_path,
                                    file_info=self.image_info_row,
                                    output=f"Decompression didn't produce a file (or it was empty)"
                                           f" {self.compression_results.original_path}")

                            measured_times.append(self._decompression_results.decompression_time_seconds)
                            if repetition_index < options.repetitions - 1:
                                os.remove(tmp_reconstructed_path)
                    self._decompression_results.decompression_time_seconds = sum(measured_times) / len(measured_times)
                except Exception as ex:
                    os.remove(tmp_reconstructed_path)
                    raise ex

            return self._decompression_results

        @property
        def numpy_dtype(self):
            """Get the numpy dtype corresponding to the original image's data format
            """

            return isets.iproperties_row_to_numpy_dtype(self.image_info_row)

        def __del__(self):
            if self._compression_results is not None:
                try:
                    os.remove(self._compression_results.compressed_path)
                except OSError:
                    pass
            if self._decompression_results is not None:
                try:
                    os.remove(self._decompression_results.reconstructed_path)
                except OSError:
                    pass

    def __init__(self, codecs,
                 dataset_paths=None,
                 csv_experiment_path=None,
                 csv_dataset_path=None,
                 dataset_info_table=None,
                 overwrite_file_properties=False,
                 reconstructed_dir_path=None,
                 compressed_copy_dir_path=None,
                 task_families=None):
        """
        :param codecs: list of :py:class:`AbstractCodec` instances. Note that
          codecs are compatible with the interface of :py:class:`ExperimentTask`.
        :param dataset_paths: list of paths to the files to be used as input for compression.
          If it is None, this list is obtained automatically from the configured
          base dataset dir.
        :param csv_experiment_path: if not None, path to the CSV file giving persistence
          support to this experiment.
          If None, it is automatically determined within options.persistence_dir.
        :param csv_dataset_path: if not None, path to the CSV file given persistence
          support to the dataset file properties.
          If None, it is automatically determined within options.persistence_dir.
        :param dataset_info_table: if not None, it must be a ImagePropertiesTable instance or
          subclass instance that can be used to obtain dataset file metainformation,
          and/or gather it from csv_dataset_path. If None, a new ImagePropertiesTable
          instance is created and used for this purpose.
        :param overwrite_file_properties: if True, file properties are recomputed before starting
          the experiment. Useful for temporary and/or random datasets. Note that overwrite
          control for the experiment results themselves is controlled in the call
          to get_df
        :param reconstructed_dir_path: if not None, a directory where reconstructed images are
          to be stored.
        :param compressed_copy_dir_path: if not None, it gives the directory where a copy of the compressed images.
          is to be stored. If may not be generated for images for which all columns are known
        :param task_families: if not None, it must be a list of TaskFamily instances. It is used to set the
          "family_label" column for each row. If the codec is not found within the families, a default
          label is set indicating so.
        """
        table_class = type(dataset_info_table) if dataset_info_table is not None \
            else self.default_file_properties_table_class
        csv_dataset_path = csv_dataset_path if csv_dataset_path is not None \
            else os.path.join(options.persistence_dir, f"{table_class.__name__}_persistence.csv")
        imageinfo_table = dataset_info_table if dataset_info_table is not None \
            else table_class(csv_support_path=csv_dataset_path)

        csv_dataset_path = csv_dataset_path if csv_dataset_path is not None \
            else f"{dataset_info_table.__class__.__name__}_persistence.csv"

        if not codecs:
            raise ValueError(f"{self}: no codecs were selected for this experiment. At least one is needed.")
        non_subclass_codecs = [c for c in codecs if not isinstance(c, enb.icompression.AbstractCodec)]
        if non_subclass_codecs:
            enb.logger.warn(f"Compression experiment {self.__class__.__name__} received parameter codecs "
                            f"with {len(non_subclass_codecs)} objects not inheriting from "
                            f"enb.icompression.AbstractCodec: "
                            f"{', '.join(repr(c) for c in non_subclass_codecs)}.\n"
                            f"You can remove this warning by explicitly inheriting from that class for "
                            f"the aforementioned instances.")

        super().__init__(tasks=codecs,
                         dataset_paths=dataset_paths,
                         csv_experiment_path=csv_experiment_path,
                         csv_dataset_path=csv_dataset_path,
                         dataset_info_table=imageinfo_table,
                         overwrite_file_properties=overwrite_file_properties,
                         task_families=task_families)
        self.reconstructed_dir_path = reconstructed_dir_path
        self.compressed_copy_dir_path = compressed_copy_dir_path
        # This attribute is automatically set before running the defined column-setting functions,
        # then set back to None after that. It enables lazy and at-most-once compression/decompression.
        self.codec_results = None

    @property
    def codecs(self):
        """:return: an iterable of defined codecs
        """
        return self.tasks_by_name.values()

    @codecs.setter
    def codecs(self, new_codecs):
        self.tasks_by_name = collections.OrderedDict({
            codec.name: codec for codec in new_codecs})

    @property
    def codecs_by_name(self):
        """Alias for :py:attr:`tasks_by_name`
        """
        return self.tasks_by_name

    def compute_one_row(self, filtered_df, index, loc, column_fun_tuples, overwrite):

        # Prepare a new column with a self.CodecRowWrapper instance that allows
        # automatic, lazy computation of compression and decompression results.
        file_path, codec_name = index
        codec = self.codecs_by_name[codec_name]
        image_info_row = self.dataset_table_df.loc[indices_to_internal_loc(file_path)]
        
        # A temporary attribute is created with a self.CompressionDecompressionWrapper instance,
        # which allows lazy, at-most-one execution of the compression/decompression process.
        # Column-setting methods can access the wrapper with self.
        try:
            assert self.codec_results is None
        except AttributeError:
            pass
        try:
            self.codec_results = self.CompressionDecompressionWrapper(
                file_path=file_path, codec=codec,
                image_info_row=image_info_row)
            assert self.codec_results is not None

            processed_row = super().compute_one_row(
                filtered_df=filtered_df, index=index, loc=loc, column_fun_tuples=column_fun_tuples,
                overwrite=overwrite)

            if isinstance(processed_row, Exception):
                # Should not do anything beyond here if errors occurred
                return processed_row

            if self.compressed_copy_dir_path:
                output_compressed_path = os.path.join(
                    self.compressed_copy_dir_path,
                    codec.name,
                    os.path.basename(os.path.dirname(file_path)), os.path.basename(file_path))
                os.makedirs(os.path.dirname(output_compressed_path), exist_ok=True)
                if options.verbose > 1:
                    print(f"[C]opying {file_path} into {output_compressed_path}")
                shutil.copy(self.codec_results.compression_results.compressed_path, output_compressed_path)

            if self.reconstructed_dir_path is not None:
                output_reconstructed_path = os.path.join(
                    self.reconstructed_dir_path,
                    codec.name,
                    os.path.basename(os.path.dirname(file_path)), os.path.basename(file_path))
                os.makedirs(os.path.dirname(output_reconstructed_path), exist_ok=True)
                if options.verbose > 1:
                    print(
                        f"[C]opying {self.codec_results.compression_results.compressed_path} into {output_reconstructed_path}")
                shutil.copy(self.codec_results.decompression_results.reconstructed_path,
                            output_reconstructed_path)

                if image_info_row["component_count"] == 3:
                    rendered_path = f"{output_reconstructed_path}.png"
                    if not os.path.exists(rendered_path) or options.force:
                        array = isets.load_array_bsq(
                            file_or_path=self.codec_results.decompression_results.reconstructed_path,
                            image_properties_row=image_info_row).astype(np.int)
                        cmin = array.min()
                        cmax = array.max()
                        array = np.round((255 * (array.astype(np.int) - cmin) / (cmax - cmin))).astype("uint8")
                        if options.verbose > 1:
                            print(f"[R]endering {rendered_path}")

                        numpngw.imwrite(rendered_path, array.swapaxes(0, 1))

                else:
                    full_array = isets.load_array_bsq(
                        file_or_path=self.codec_results.decompression_results.reconstructed_path,
                        image_properties_row=image_info_row).astype(np.int)
                    for i, rendered_path in enumerate(f"{output_reconstructed_path}_component{i}.png"
                                                      for i in range(image_info_row['component_count'])):
                        if not os.path.exists(rendered_path) or options.force:
                            array = full_array[:, :, i].squeeze().swapaxes(0, 1)
                            cmin = array.min()
                            cmax = array.max()
                            array = np.round((255 * (array - cmin) / (cmax - cmin))).astype("uint8")
                            if options.verbose > 1:
                                print(f"[R]endering {rendered_path}")
                            numpngw.imwrite(rendered_path, array)
        finally:
            del self.codec_results
            self.codec_results = None

            try:
                del self.codec_results
            except AttributeError:
                pass


        return processed_row

    @atable.column_function("compressed_size_bytes", label="Compressed data size (Bytes)", plot_min=0)
    def set_compressed_data_size(self, index, row):
        row[_column_name] = os.path.getsize(self.codec_results.compression_results.compressed_path)

    @atable.column_function([
        atable.ColumnProperties(name="compression_ratio", label="Compression ratio", plot_min=0),
        atable.ColumnProperties(name="lossless_reconstruction", label="Lossless?"),
        atable.ColumnProperties(name="compression_time_seconds", label="Compression time (s)", plot_min=0),
        atable.ColumnProperties(name="decompression_time_seconds", label="Decompression time (s)", plot_min=0),
        atable.ColumnProperties(name="repetitions", label="Number of compression/decompression repetitions",
                                plot_min=0),
        atable.ColumnProperties(name="compressed_file_sha256", label="Compressed file's SHA256")
    ])
    def set_comparison_results(self, index, row):
        """Perform a compression-decompression cycle and store the comparison results
        """
        file_path, codec_name = index
        row.image_info_row = self.dataset_table_df.loc[indices_to_internal_loc(file_path)]
        assert self.codec_results.compression_results.compressed_path == self.codec_results.decompression_results.compressed_path
        try:
            assert row.image_info_row["bytes_per_sample"] * row.image_info_row["samples"] \
                   == os.path.getsize(self.codec_results.compression_results.original_path)
        except (KeyError, AssertionError) as ex:
            enb.logger.debug(f"Could not verify valid size. {repr(ex)}")
        hasher = hashlib.sha256()
        with open(self.codec_results.compression_results.compressed_path, "rb") as compressed_file:
            hasher.update(compressed_file.read())
        compressed_file_sha256 = hasher.hexdigest()

        row["lossless_reconstruction"] = filecmp.cmp(self.codec_results.compression_results.original_path,
                                                     self.codec_results.decompression_results.reconstructed_path)
        assert self.codec_results.compression_results.compression_time_seconds is not None
        row["compression_time_seconds"] = self.codec_results.compression_results.compression_time_seconds
        assert self.codec_results.decompression_results.decompression_time_seconds is not None
        row["decompression_time_seconds"] = self.codec_results.decompression_results.decompression_time_seconds
        row["repetitions"] = options.repetitions
        row["compression_ratio"] = os.path.getsize(self.codec_results.compression_results.original_path) / row[
            "compressed_size_bytes"]
        row["compressed_file_sha256"] = compressed_file_sha256

    @atable.column_function("bpppc", label="Compressed data rate (bpppc)", plot_min=0)
    def set_bpppc(self, index, row):
        try:
            row[_column_name] = 8 * row["compressed_size_bytes"] / row.image_info_row["samples"]
        except KeyError as ex:
            enb.logger.debug(f"Cannot determine bpppc: {repr(ex)}")
            assert "compressed_size_bytes" in row

    @atable.column_function("compression_ratio_dr", label="Compression ratio", plot_min=0)
    def set_compression_ratio_dr(self, index, row):
        """Set the compression ratio calculated based on the dynamic range of the
        input samples, as opposed to 8*bytes_per_sample.
        """
        row[_column_name] = (row.image_info_row["dynamic_range_bits"] * row.image_info_row["samples"]) \
                            / (8 * row["compressed_size_bytes"])

    @atable.column_function(
        [atable.ColumnProperties(name="compression_efficiency_1byte_entropy",
                                 label="Compression efficiency (1B entropy)", plot_min=0),
         atable.ColumnProperties(name="compression_efficiency_2byte_entropy",
                                 label="Compression efficiency (2B entropy)", plot_min=0)])
    def set_efficiency(self, index, row):
        compression_efficiency_1byte_entropy = \
            row.image_info_row["entropy_1B_bps"] * row.image_info_row["size_bytes"] \
            / (row["compressed_size_bytes"] * 8)
        compression_efficiency_2byte_entropy = \
            row.image_info_row["entropy_2B_bps"] * (row.image_info_row["size_bytes"] / 2) \
            / (row["compressed_size_bytes"] * 8)
        row["compression_efficiency_1byte_entropy"] = compression_efficiency_1byte_entropy
        row["compression_efficiency_2byte_entropy"] = compression_efficiency_2byte_entropy


class LosslessCompressionExperiment(CompressionExperiment):
    @atable.redefines_column
    def set_comparison_results(self, index, row):
        super().set_comparison_results(index=index, row=row)
        if not row["lossless_reconstruction"]:
            raise CompressionException(
                original_path=index[0], file_info=row,
                output="Failed to produce lossless compression for "
                       f"{index[0]} and {index[1]}")


class LossyCompressionExperiment(CompressionExperiment):
    @atable.column_function("mse", label="MSE", plot_min=0)
    def set_MSE(self, index, row):
        """Set the mean squared error of the reconstructed image.
        """
        original_array = np.fromfile(self.codec_results.compression_results.original_path,
                                     dtype=self.codec_results.numpy_dtype).astype(np.int64)

        reconstructed_array = np.fromfile(self.codec_results.decompression_results.reconstructed_path,
                                          dtype=self.codec_results.numpy_dtype).astype(np.int64)
        row[_column_name] = np.average(((original_array - reconstructed_array) ** 2))

    @atable.column_function("pae", label="PAE", plot_min=0)
    def set_PAE(self, index, row):
        """Set the peak absolute error (maximum absolute pixelwise difference)
        of the reconstructed image.
        """
        original_array = np.fromfile(self.codec_results.compression_results.original_path,
                                     dtype=self.codec_results.numpy_dtype).astype(np.int64)

        reconstructed_array = np.fromfile(self.codec_results.decompression_results.reconstructed_path,
                                          dtype=self.codec_results.numpy_dtype).astype(np.int64)
        row[_column_name] = np.max(np.abs(original_array - reconstructed_array))

    @atable.column_function("psnr_bps", label="PSNR (dB)", plot_min=0)
    def set_PSNR_nominal(self, index, row):
        """Set the PSNR assuming nominal dynamic range given by bytes_per_sample.
        """
        path, task = self.index_to_path_task(index)

        if row.image_info_row["float"]:
            # TODO: use the float type dynamic range?
            row[_column_name] = float("inf")
        else:
            max_error = (2 ** (8 * row.image_info_row["bytes_per_sample"])) - 1
            row[_column_name] = (20 * math.log10((max_error) / math.sqrt(row["mse"]))) \
                if row["mse"] > 0 else float("inf")

    @atable.column_function("psnr_dr", label="PSNR (dB)", plot_min=0)
    def set_PSNR_dynamic_range(self, index, row):
        """Set the PSNR assuming dynamic range given by dynamic_range_bits.
        """
        max_error = (2 ** row.image_info_row["dynamic_range_bits"]) - 1
        row[_column_name] = 20 * math.log10(max_error / math.sqrt(row["mse"])) \
            if row["mse"] > 0 else float("inf")


class StructuralSimilarity(CompressionExperiment):
    """Set the Structural Similarity (SSIM) and Multi-Scale Structural Similarity metrics (MS-SSIM)
    to measure the similarity between two images.

    Authors:
        - http://www.cns.nyu.edu/~lcv/ssim/msssim.zip
        - https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py
    """

    @atable.column_function([
        atable.ColumnProperties(name="ssim", label="SSIM", plot_max=1),
        atable.ColumnProperties(name="ms_ssim", label="MS-SSIM", plot_max=1)])
    def set_StructuralSimilarity(self, index, row):

        # TODO: Would it be possible to call to isets.load_array_bsq and change axes as necessary?
        original_array = np.fromfile(self.codec_results.compression_results.original_path,
                                     dtype=self.codec_results.numpy_dtype)
        original_array = np.reshape(original_array,
                                    (row.image_info_row["width"], row.image_info_row["height"],
                                     row.image_info_row["component_count"]))

        reconstructed_array = np.fromfile(self.codec_results.decompression_results.reconstructed_path,
                                          dtype=self.codec_results.numpy_dtype)
        reconstructed_array = np.reshape(reconstructed_array,
                                         (row.image_info_row["width"], row.image_info_row["height"],
                                          row.image_info_row["component_count"]))

        row["ssim"] = self.compute_SSIM(original_array, reconstructed_array)
        row["ms_ssim"] = self.cumpute_MSSIM(original_array, reconstructed_array)

    def cumpute_MSSIM(self, img1, img2, max_val=255, filter_size=11, filter_sigma=1.5, k1=0.01, k2=0.03, weights=None):
        """Return the MS-SSIM score between `img1` and `img2`.

        This function implements Multi-Scale Structural Similarity (MS-SSIM) Image
        Quality Assessment according to Zhou Wang's paper, "Multi-scale structural
        similarity for image quality assessment" (2003).
        Link: https://ece.uwaterloo.ca/~z70wang/publications/msssim.pdf

        Author's MATLAB implementation:
        http://www.cns.nyu.edu/~lcv/ssim/msssim.zip

        Author's Python implementation:
        https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py

        Authors documentation:

        :param img1: Numpy array holding the first RGB image batch.
        :param img2: Numpy array holding the second RGB image batch.
        :param max_val: the dynamic range of the images (i.e., the difference between the
                maximum the and minimum allowed values).
        :param filter_size: Size of blur kernel to use (will be reduced for small
              images).
        :param filter_sigma: Standard deviation for Gaussian blur kernel (will be reduced
                for small images).
        :param k1: Constant used to maintain stability in the SSIM calculation (0.01 in
                the original paper).
        :param k2: Constant used to maintain stability in the SSIM calculation (0.03 in
                the original paper).
        """
        if img1.shape != img2.shape:
            raise RuntimeError('Input images must have the same shape (%s vs. %s).',
                               img1.shape, img2.shape)
        if img1.ndim != 3:
            raise RuntimeError('Input images must have four dimensions, not %d',
                               img1.ndim)

        weights = np.array(weights if weights else
                           [0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
        levels = weights.size
        downsample_filter = np.ones((2, 2, 1)) / 4.0
        im1, im2 = [x.astype(np.float64) for x in [img1, img2]]
        mssim = np.array([])
        mcs = np.array([])
        for _ in range(levels):
            ssim, cs = self.compute_SSIM(
                im1, im2, max_val=max_val, filter_size=filter_size,
                filter_sigma=filter_sigma, k1=k1, k2=k2, full=True)
            mssim = np.append(mssim, ssim)
            mcs = np.append(mcs, cs)
            filtered = [convolve(im, downsample_filter, mode='reflect')
                        for im in [im1, im2]]
            im1, im2 = [x[::2, ::2, :] for x in filtered]

        return np.prod(mcs[0:levels - 1] ** weights[0:levels - 1]) * (mssim[levels - 1] ** weights[levels - 1])

    def compute_SSIM(self, img1, img2, max_val=255, filter_size=11, filter_sigma=1.5, k1=0.01, k2=0.03, full=False):
        """Return the Structural Similarity Map between `img1` and `img2`.

        This function attempts to match the functionality of ssim_index_new.m by
        Zhou Wang: http://www.cns.nyu.edu/~lcv/ssim/msssim.zip

        Author's Python implementation:
        https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py

        :param img1: Numpy array holding the first RGB image batch.
        :param img2: Numpy array holding the second RGB image batch.
        :param max_val: the dynamic range of the images (i.e., the difference between the
                maximum the and minimum allowed values).
        :param filter_size: Size of blur kernel to use (will be reduced for small
              images).
        :param filter_sigma: Standard deviation for Gaussian blur kernel (will be reduced
                for small images).
        :param k1: Constant used to maintain stability in the SSIM calculation (0.01 in
                the original paper).
        :param k2: Constant used to maintain stability in the SSIM calculation (0.03 in
                the original paper).
        """
        if img1.shape != img2.shape:
            raise RuntimeError('Input images must have the same shape (%s vs. %s).',
                               img1.shape, img2.shape)
        if img1.ndim != 3:
            raise RuntimeError('Input images must have four dimensions, not %d',
                               img1.ndim)
        img1 = img1.astype(np.float64)
        img2 = img2.astype(np.float64)
        height, width, bytes = img1.shape

        # Filter size can't be larger than height or width of images.
        size = min(filter_size, height, width)

        # Scale down sigma if a smaller filter size is used.
        sigma = size * filter_sigma / filter_size if filter_size else 0

        if filter_size:
            window = np.reshape(self._FSpecialGauss(size, sigma), (size, size, 1))
            mu1 = signal.fftconvolve(img1, window, mode='valid')
            mu2 = signal.fftconvolve(img2, window, mode='valid')
            sigma11 = signal.fftconvolve(img1 * img1, window, mode='valid')
            sigma22 = signal.fftconvolve(img2 * img2, window, mode='valid')
            sigma12 = signal.fftconvolve(img1 * img2, window, mode='valid')
        else:
            # Empty blur kernel so no need to convolve.
            mu1, mu2 = img1, img2
            sigma11 = img1 * img1
            sigma22 = img2 * img2
            sigma12 = img1 * img2

        mu11 = mu1 * mu1
        mu22 = mu2 * mu2
        mu12 = mu1 * mu2
        sigma11 -= mu11
        sigma22 -= mu22
        sigma12 -= mu12

        # Calculate intermediate values used by both ssim and cs_map.
        c1 = (k1 * max_val) ** 2
        c2 = (k2 * max_val) ** 2
        v1 = 2.0 * sigma12 + c2
        v2 = sigma11 + sigma22 + c2
        ssim = np.mean((((2.0 * mu12 + c1) * v1) / ((mu11 + mu22 + c1) * v2)))
        cs = np.mean(v1 / v2)
        if full:
            return ssim, cs
        return ssim

    def _FSpecialGauss(self, size, sigma):

        radius = size // 2
        offset = 0.0
        start, stop = -radius, radius + 1
        if size % 2 == 0:
            offset = 0.5
            stop -= 1
        x, y = np.mgrid[offset + start:stop, offset + start:stop]
        assert len(x) == size
        g = np.exp(-((x ** 2 + y ** 2) / (2.0 * sigma ** 2)))
        return g / g.sum()


class SpectralAngleTable(LossyCompressionExperiment):
    """Lossy compression experiment that computes spectral angle "distance"
    measures between the compressed and the reconstructed images.
    
    Subclasses of LossyCompressionExperiment may inherit from this one to
    automatically add the data columns defined here
    """

    def get_spectral_angles_deg(self, index, row):
        """Return a sequence of spectral angles (in degrees), 
        one per (x,y) position in the image, flattened in raster order.
        """
        # Read original and reconstructed images
        original_file_path, task_name = index
        image_properties_row = self.get_dataset_info_row(original_file_path)
        decompression_results = self.codec_results.decompression_results
        original_array = isets.load_array_bsq(
            file_or_path=original_file_path, image_properties_row=image_properties_row)
        reconstructed_array = isets.load_array_bsq(
            file_or_path=decompression_results.reconstructed_path, image_properties_row=image_properties_row)

        # Reshape flattening the x,y axes, and maintaining the z axis for each (x,y) position
        original_array = np.reshape(
            original_array.swapaxes(0, 1),
            (image_properties_row["width"] * image_properties_row["height"], image_properties_row["component_count"]),
            "F").astype("i8")
        reconstructed_array = np.reshape(
            reconstructed_array.swapaxes(0, 1),
            (image_properties_row["width"] * image_properties_row["height"], image_properties_row["component_count"]),
            "F").astype("i8")

        dots = np.einsum("ij,ij->i", original_array, reconstructed_array)
        magnitude_a = np.linalg.norm(original_array, axis=1)
        magnitude_b = np.linalg.norm(reconstructed_array, axis=1)

        for i in range(magnitude_a.shape[0]):
            # Avoid division by zero
            magnitude_a[i] = max(1e-4, magnitude_a[i])
            magnitude_b[i] = max(1e-4, magnitude_b[i])

        # Clip, because the dot product can slip past 1 or -1 due to rounding
        cosines = np.clip(dots / (magnitude_a * magnitude_b), -1, 1)
        angles = np.degrees(np.arccos(cosines))
        # Round because two identical images should return an angle of exactly 0
        angles = np.round(angles, 5)

        return angles.tolist()

    @atable.column_function([
        enb.atable.ColumnProperties("mean_spectral_angle_deg", label="Mean spectral angle (deg)",
                                    plot_min=0, plot_max=None),
        enb.atable.ColumnProperties("max_spectral_angle_deg", label="Max spectral angle (deg)",
                                    plot_min=0, plot_max=None)])
    def set_spectral_distances(self, index, row):
        spectral_angles = self.get_spectral_angles_deg(index=index, row=row)

        for angle in spectral_angles:
            assert not np.isnan(angle), f"Error calculating an angle for {index}: {angle}"

        row["mean_spectral_angle_deg"] = sum(spectral_angles) / len(spectral_angles)
        row["max_spectral_angle_deg"] = max(spectral_angles)



#!/usr/bin/env python3
"""Tools define plugin installables.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import os
import glob
import inspect
import shutil
import platform
import subprocess

from .installable import Installable


class Plugin(Installable):
    """Plugins are self-contained, python modules that may assume enb is installed.

    - They can be installed into your projects via the enb CLI (e.g., with `enb plugin install <name> <clone_dir>`),
      and then imported like any other module.

    - The list of plugins available off-the-box can be obtained using the enb CLI (e.g., with `enb plugin list`).

    - Plugins may declare pip dependencies, which are attempted to be satisfied automatically when the plugins
      are installed. In addition, plugins may define their `extra_requirements_message` member to be not None,
      in which case it describes manual intervention required from the user either as a pre-installation
      or post-installation step. That message is shown when attempting to install the plugin with
      not-None `extra_requirements_message`.

    - The __init__.py file is preserved if present. If not present, one is created automatically,
      which imports all symbols from all .py modules in the plugin.
    """

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False):
        """Make a copy of this plugin into installation_dir, ready to be imported.
        By default, a verbatim copy of the source plugin's dir is made.
        Any previous contents in installation_dir are overwritten.
        Then any explicit requirements are met (external software may be downloaded
        and pip packages installed).

        :param overwrite_destination: if True, the destination path is deleted before
          installation. If False and installation_dir already exists, an error
          is raised (plugins are intended to be self-contained, isolated python modules).
        """
        if not overwrite_destination and os.path.exists(installation_dir):
            raise ValueError(f"Plugin {repr(cls)} cannot be installed into existing "
                             f"path {installation_dir} because "
                             f"overwrite_destination={overwrite_destination}.")
        super().install(installation_dir=installation_dir,
                        overwrite_destination=True)
        print(f"Plugin {repr(cls.name)} successfully installed into {repr(installation_dir)}.")

    @classmethod
    def build(cls, installation_dir):
        """Perform any additional retrieval, compilation and setup necessary for this plugin
        to be importable and usable. By default:

        - The existence of installation_dir as a directory is performed.
        - Install any required apt modules
        - Any needed python modules are installed via pip
        - Cleanup any generic files that might not be needed at this point
        - The __init__.py file is preserved or generated automatically
        """
        # Existence assertion
        assert os.path.isdir(installation_dir), \
            f"{cls.__name__}.build(installation_dir={repr(installation_dir)}): installation_dir does not exist"

        # cleanup
        shutil.rmtree(os.path.join(installation_dir, "__pycache__"), ignore_errors=True)

        # add custom __init__ if needed
        init_path = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(cls))), "__init__.py")
        with open(os.path.join(installation_dir, "__init__.py"), "w") as init_file:
            if os.path.exists(init_path):
                # An __init__.py file already existed, simply copy it
                with open(init_path, "r") as source_init:
                    init_file.write(source_init.read())
            else:
                # No __init__.py found. Generating one that publishes all symbols by default
                for py_path in [p for p in glob.glob(os.path.join(installation_dir, "*.py"))
                                if not os.path.basename(p).startswith("__")]:
                    module_name = os.path.basename(py_path)[:-3]
                    init_file.write(f"from . import {module_name}\n")
                    init_file.write(f"from .{module_name} import *\n\n")


class PluginMake(Plugin):
    """Plugin that assumes the existence of a valid Makefile in the installation folder,
    and uses it for building the plugin.
    """

    @classmethod
    def build(cls, installation_dir):
        super().build(installation_dir=installation_dir)
        platform_name = platform.system().lower()
        make_path = os.path.join(installation_dir, f"Makefile.{platform_name}")
        if not os.path.exists(make_path):
            make_path = os.path.join(installation_dir, f"Makefile")
        if os.path.exists(make_path):
            print(f"Building downloaded plugin {cls.name}...")
            invocation = f"cd {os.path.dirname(os.path.abspath(make_path))} && make -f {os.path.basename(make_path)}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Error bulding {repr(cls)} with {make_path}. "
                                f"Status = {status} != 0."
                                f"\nInput=[{invocation}].\nOutput=[{output}]")
        else:
            raise ValueError(f"Cannot build {repr(cls)}: no valid makefile "
                             f"in {installation_dir}.")



#!/usr/bin/env python3
"""Unit tests for isets.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/07"

import unittest
import tempfile
import os
import numpy as np

import test_all
from enb import pgm
from enb import isets


class TestPGM(unittest.TestCase):

    def test_pgm_read_write(self):
        width = 67
        height = 34

        for bytes_per_sample in [1, 2]:
            array = np.zeros((width, height), dtype=f"u{bytes_per_sample}")
            max_pixel_value = 2 ** (8 * bytes_per_sample) - 1
            i = 0
            for y in range(height):
                for x in range(width):
                    array[x, y] = (i % (max_pixel_value + 1))
                    i += 1
            array[0, 0] = max_pixel_value
            assert array[0, 0] == max_pixel_value

            with tempfile.NamedTemporaryFile() as tmp_file:
                pgm.write_pgm(array=array, bytes_per_sample=bytes_per_sample, output_path=tmp_file.name)
                loaded_array = pgm.read_pgm(input_path=tmp_file.name)
                assert array.shape == loaded_array.shape
                assert (array == loaded_array).all(), (array, '\n', loaded_array)


if __name__ == '__main__':
    unittest.main()



import enb.plugins


class MCALICPlugin(enb.plugins.Plugin):
    name = "mcalic"
    label = "Wraper for E. Magli et al.'s M-CALIC codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Enrico Magli et al."]
    contrib_reference_urls = ["https://doi.org/10.1109/LGRS.2003.822312"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/mcalic/Mcalic_enc_nl?raw=true",
         "Mcalic_enc_nl"),
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/mcalic/Mcalic_dec_nl?raw=true",
         "Mcalic_dec_nl"),
    ]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Huffman codec wrapper
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
from enb import icompression


class Huffman(icompression.WrapperCodec, icompression.LosslessCodec):
    def __init__(self, huff_binary=os.path.join(os.path.dirname(os.path.abspath(__file__)), "fse")):
        icompression.WrapperCodec.__init__(self,
                                           compressor_path=huff_binary,
                                           decompressor_path=huff_binary,
                                           param_dict=None, output_invocation_dir=None)

        self.huff_binary = huff_binary

        assert os.path.isfile(huff_binary)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-h -f {original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-h -d -f {compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return "Huffman"



import enb.plugins


class LZBZPlugin(enb.plugins.Plugin):
    name = "zip"
    label = "Assortment of lz-based and bzip-based codecs"
    tags = {"data compression", "codec"}
    contrib_authors = ["The zlib team", "The lzma team", "The bz2 team"]
    contrib_reference_urls = ["https://docs.python.org/3/library/zlib.html",
                              "https://docs.python.org/3/library/lzma.html",
                              "https://docs.python.org/3/library/bz2.html"]
    tested_on = {"linux"}



import enb


class PortExperimentTemplate(enb.plugins.Template):
    """Self-contained example using the basic usage of enb's experiment class
    """
    name = "port-experiment-example"
    author = ["Miguel Hernández-Cabronero"]
    label = "Self-contained port scanning experiment example"
    tags = {"documentation"}
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Codecs wrapping V2F forests.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/09/01"

import os
import subprocess
import enb

# Constants taken from the C code for consistency and maintainability.
V2F_C_QUANTIZER_MODE_NONE = 0
V2F_C_QUANTIZER_MODE_UNIFORM = 1
V2F_C_QUANTIZER_MODE_COUNT = 2
V2F_C_DECORRELATOR_MODE_NONE = 0
V2F_C_DECORRELATOR_MODE_LEFT = 1
V2F_C_DECORRELATOR_MODE_2_LEFT = 2
V2F_C_DECORRELATOR_MODE_JPEGLS = 3
V2F_C_DECORRELATOR_MODE_COUNT = 4
V2F_C_QUANTIZER_MODE_MAX_STEP_SIZE = 255


class V2FCodec(enb.icompression.LosslessCodec,
               enb.icompression.NearLosslessCodec,
               enb.icompression.WrapperCodec):
    """Wrapper for the V2F compressor/decompressor tools in C,
    which must be present in the same path as this script.
    """

    def __init__(self, v2fc_header_path, qstep=None, quantizer_mode=None, decorrelator_mode=None,
                 verify_on_initialization=True, time_results_dir=None):
        """
        Initialize a V2F codec that employs the forest definition in `v2fc_header_path`, and optionally
        overwrite the quantization and decorrelation defined in them.

        :param v2fc_header_path: path to the V2F forest definition to be employed
        :param qstep: quantization step size. Only applied if quantizer_mode is not
          V2F_C_DECORRELATOR_MODE_NONE. Cannot exceed V2F_C_QUANTIZER_MODE_MAX_STEP_SIZE.
        :param quantizer_mode: quantization mode; see the V2F_C_QUANTIZER_MODE_* constants.
        :param decorrelator_mode: decorrelation (e.g., prediction) mode;
          see see the V2F_C_DECORRELATOR_MODE_* constants.
        :param verify_on_initialization: if True, the codec in v2fc_heder_path is verified upon initialization.
        :param time_results_dir: if not None, it must be the path of a directory that will be used to store time
          measurements reported by the V2F prototype.
        """
        self.codec_root_dir = os.path.abspath(os.path.dirname(__file__))
        super().__init__(compressor_path=os.path.join(self.codec_root_dir, "v2f_compress"),
                         decompressor_path=os.path.join(self.codec_root_dir, "v2f_decompress"),
                         param_dict=dict(
                             v2fc_header_path=v2fc_header_path,
                             qstep=qstep,
                             quantizer_mode=quantizer_mode,
                             decorrelator_mode=decorrelator_mode))

        # Header path with the V2F codec (including forest) definition
        self.v2fc_header_path = v2fc_header_path
        if not self.v2fc_header_path or not os.path.isfile(self.v2fc_header_path):
            raise ValueError(f"Cannot find a V2F codec definition at {repr(self.v2fc_header_path)}. "
                             f"Please make sure you are passing the right argument to "
                             f"{self.__class__}'s initialzier, and that the file is actually present.")

        # Verify the chosen codec if configured to do so.
        self.verifier_path = os.path.join(self.codec_root_dir, "v2f_verify_codec")
        if verify_on_initialization:
            self.verify_codec(v2fc_header_path=v2fc_header_path)

        # Prepare the folder where time measurements are to be stored
        self.time_results_dir = time_results_dir
        if self.time_results_dir:
            os.makedirs(self.time_results_dir, exist_ok=True)

    def verify_codec(self, v2fc_header_path):
        """Run the V2F codec verification tool for a given path.
        An exception is raised if this verification fails.
        """
        if not os.path.isfile(self.verifier_path):
            raise ValueError(f"Could not find the V2F codec verifier tool at {repr(self.verifier_path)}. "
                             f"Please make sure it is present at that path and try again.")

        with enb.logger.info_context(f"Verifying V2F codec {repr(v2fc_header_path)}"):
            invocation = f"{self.verifier_path} {os.path.abspath(v2fc_header_path)}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise ValueError(f"V2F codec verification failed for {repr(v2fc_header_path)}.\n"
                                 f"status: {status}\n"
                                 f"invocation: {repr(invocation)}\n"
                                 f"execution output: {output}.")
            else:
                enb.logger.info(f"Verification OK. Returned message:\n{output}")

    def get_time_path(self, original_path):
        assert self.time_results_dir, f"Trying to run get_time_path() without having set self.time_results_dir."
        return os.path.join(self.time_results_dir,
                            f"q{self.param_dict['quantizer_mode']}_"
                            f"s{self.param_dict['qstep']}_"
                            f"d{self.param_dict['decorrelator_mode']}_" + \
                            os.path.basename(self.v2fc_header_path).replace(os.sep, "__"),
                            os.path.abspath(original_path).replace(os.sep, "__")
                            + "_times.csv")

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_path} {self.get_optional_argument_string(original_path=original_path)} " \
               f"{self.v2fc_header_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {self.get_optional_argument_string(original_path=None)} " \
               f"{self.v2fc_header_path} {reconstructed_path}"

    def get_optional_argument_string(self, original_path):
        optional_str = ""
        if self.param_dict['quantizer_mode'] is not None:
            optional_str += f"-q {self.param_dict['quantizer_mode']} "
        if self.param_dict['qstep'] is not None:
            optional_str += f"-s {self.param_dict['qstep']} "
        if self.param_dict['decorrelator_mode'] is not None:
            optional_str += f"-d {self.param_dict['decorrelator_mode']} "
        if self.time_results_dir is not None and original_path is not None:
            time_path = self.get_time_path(original_path)
            os.makedirs(os.path.dirname(time_path), exist_ok=True)
            optional_str += f"-t {time_path} "
        return optional_str

    @property
    def label(self):
        if self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_LEFT:
            dec_str = "left"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_2_LEFT:
            dec_str = "two left"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_JPEGLS:
            dec_str = "JPEG-LS"
        else:
            raise ValueError(self.param_dict["decorrelator_mode"])

        return f"V2F Qstep {self.param_dict['qstep']}, {dec_str}"



import enb.plugins


class LZ4Plugin(enb.plugins.PluginMake):
    name = "lz4"
    label = "Wraper for a LZ4 codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Yann Collet"]
    contrib_reference_urls = ["https://lz4.github.io/lz4/"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/lz4-1.9.2.zip?raw=true",
         "lz4-1.9.2.zip")]
    tested_on = {"linux"}
